from substance.exceptions import (SubstanceDriverError)


class VirtualBoxError(SubstanceDriverError):
    '''
    Virtual box driver raised error
    '''
    code = None

    def __init__(self, code=None, message=''):
        super(VirtualBoxError, self).__init__(message)
        self.code = code


class VirtualBoxMissingAdditions(SubstanceDriverError):
    '''
    The Virtual Box guest additions are missing.
    '''


class VirtualBoxVersionError(SubstanceDriverError):
    '''
    Wrong version of the Virtual Box system is installed.
    '''


class VirtualBoxObjectNotFound(SubstanceDriverError):
    '''
    Tried to read unknown object
    '''



from substance.monads import *
from substance.logs import *
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)


class Launch(Command):

    def getUsage(self):
        return "substance launch"

    def getHelpTitle(self):
        return "Launch the current engine's virtual machine"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):

        engine = self.parent.getOption('engine')
        if engine:
            engine = OK(engine)
        else:
            engine = self.core.readCurrentEngineName()

        return engine \
            .bind(self.core.loadEngine) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.launch) \
            .catch(self.exitError)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate


class Restart(Command):
    def getShellOptions(self, optparser):
        optparser.add_option("-t", "--time", dest="time",
                             help="Seconds to wait before sending SIGKILL", default=10)
        return optparser

    def getUsage(self):
        return "substance restart [options] [CONTAINER...]"

    def getHelpTitle(self):
        return "Restart all or specified container(s)"

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envRestart, time=self.getOption('time'), containers=self.args) \
            .catch(self.exitError)



import unittest
import os
import tempfile
import tests
import re
from substance.core import Core
from substance.engine import (EngineProfile, Engine)
from substance.shell import Shell
from substance.monads import *
from substance.constants import *

import substance.driver.virtualbox.vbox as vbox
import substance.driver.virtualbox.machine as machine
import substance.driver.virtualbox.network as network


class TestVirtualBox(tests.TestBase):

    core = None
    engine = None
    basePath = None
    projectsPath = None

    vmTest = True
    vmName = "testVbox"
    vmUUID = None

    @classmethod
    def setUpClass(cls):
        cls.basePath = cls.addTemporaryDir()
        cls.projectsPath = cls.addTemporaryDir()

    @classmethod
    def tearDownClass(cls):
        if cls.vmUUID:
            machine.terminate(cls.vmUUID)
            machine.delete(cls.vmUUID)

        if cls.basePath:
            Shell.nukeDirectory(cls.basePath).catch(cls.raiser)
        if cls.projectsPath:
            Shell.nukeDirectory(cls.projectsPath).catch(cls.raiser)

    def testReadVersion(self):
        op = vbox.readVersion()
        self.assertIsInstance(op, OK)
        self.assertTrue(re.match(r'^[0-9\.]*$', op.getOK()))

    def testReadDHCPs(self):
        op = network.readDHCPs()
        self.assertIsInstance(op, OK)

        dhcps = op.getOK()
        self.assertEqual(type(dhcps), type([]))

        for dhcp in dhcps:
            self.assertIsInstance(dhcp, network.DHCP)

    def testHostOnlyOperations(self):
        op = network.readHostOnlyInterfaces()
        self.assertIsInstance(op, OK)

        hoifs = op.getOK()
        self.assertEqual(type(hoifs), type([]))
        for hoif in hoifs:
            self.assertIsInstance(hoif, network.HostOnlyInterface)

        op = network.addHostOnlyInterface()
        self.assertIsInstance(op, OK)
        iface = op.getOK()

        options = {'ip': '10.0.123.1', 'netmask': '255.255.255.0'}
        op = network.configureHostOnlyInterface(iface, **options)
        self.assertIsInstance(op, OK)

        op = network.readHostOnlyInterface(iface)
        hoif = op.getOK()
        self.assertIsInstance(op, OK)
        self.assertIsInstance(hoif, network.HostOnlyInterface)
        self.assertEqual(hoif.name, iface)

        self.assertEqual(hoif.network.ip.format(), options['ip'])
        self.assertEqual(hoif.network.netmask.format(), options['netmask'])

        dhcpOptions = {'gateway': '10.0.123.1', 'netmask': '255.255.255.0', 'lowerIP': '10.0.123.5', 'upperIP': '10.0.123.100'}
        op = network.addDHCP(iface, **dhcpOptions)
        self.assertIsInstance(op, OK)

        op = network.readDHCP(iface)
        self.assertIsInstance(op, OK)
        dhcp = op.getOK()
        self.assertIsInstance(dhcp, network.DHCP)
        self.assertEqual(dhcp.interface, iface)
        self.assertEqual(dhcp.gateway, dhcpOptions['gateway'])
        self.assertEqual(dhcp.netmask, dhcpOptions['netmask'])
        self.assertEqual(dhcp.lowerIP, dhcpOptions['lowerIP'])
        self.assertEqual(dhcp.upperIP, dhcpOptions['upperIP'])

        op = network.removeHostOnlyInterface(hoif)
        self.assertIsInstance(op, OK)
        op = network.readHostOnlyInterface(iface)
        self.assertIsInstance(op, Fail)
        op = network.readDHCP(iface)
        self.assertIsInstance(op, Fail)

    def testReadMachines(self):
        op = machine.readMachines()
        self.assertIsInstance(op, OK)
        self.assertIsInstance(op.getOK(), dict)

    def testSequence(self):
        if not self.vmTest:
            return

        self.doImport()
        self.doTestPF()
        self.doTestSF()
        self.doReadGuestAdd()
        self.doStart()
        self.doSuspend()
        self.doStart()
        self.doTerminate()
        self.doDelete()

    def testDHCP(self):
        dhcpA = network.DHCP(1, "host", "192.168.1.1", "255.255.255.0", "192.168.1.10", "192.168.1.20", True)
        dhcpB = network.DHCP(2, "priv", "172.16.1.1", "255.255.255.0", "172.16.1.10", "172.16.1.20", True)
        dhcpC = network.DHCP(3, "long", "10.0.1.1", "255.255.255.0", "10.0.1.10", "10.0.1.20", True)
        sdhcpA = network.DHCP(1, "host", "192.168.1.1", "255.255.255.0", "192.168.1.10", "192.168.1.20", True)
        sdhcpB = network.DHCP(2, "priv", "172.16.1.1", "255.255.255.0", "172.16.1.10", "172.16.1.20", True)
        sdhcpC = network.DHCP(3, "long", "10.0.1.1", "255.255.255.0", "10.0.1.10", "10.0.1.20", True)
        dhcps = [dhcpA, dhcpB, dhcpC]
        sdhcps = [sdhcpA, sdhcpB, sdhcpC]

        self.assertEqual(dhcpA, sdhcpA)
        self.assertEqual(dhcpB, sdhcpB)
        self.assertEqual(dhcpC, sdhcpC)
        self.assertIn(sdhcpA, dhcps)
        self.assertIn(sdhcpB, dhcps)
        self.assertIn(sdhcpC, dhcps)
        self.assertIn(dhcpA, sdhcps)
        self.assertIn(dhcpB, sdhcps)
        self.assertIn(dhcpC, sdhcps)

    def testHOIFs(self):
        h1 = network.HostOnlyInterface("nic1", "00:00:00:00:00:00", "192.168.1.10", "255.255.255.0", None, None, "Up", "host")
        h2 = network.HostOnlyInterface("nic2", "0f:00:00:00:00:00", "192.168.1.11", "255.255.255.0", None, None, "Up", "host")
        h3 = network.HostOnlyInterface("nic3", "0g:00:00:00:00:00", "192.168.1.12", "255.255.255.0", None, None, "Up", "host")
        sh1 = network.HostOnlyInterface("nic1", "00:00:00:00:00:00", "192.168.1.10", "255.255.255.0", None, None, "Up", "host")
        sh2 = network.HostOnlyInterface("nic2", "0f:00:00:00:00:00", "192.168.1.11", "255.255.255.0", None, None, "Up", "host")
        sh3 = network.HostOnlyInterface("nic3", "0g:00:00:00:00:00", "192.168.1.12", "255.255.255.0", None, None, "Up", "host")

        hs = [h1, h2, h3]
        shs = [sh1, sh2, sh3]

        self.assertEqual(h1, sh1)
        self.assertEqual(h2, sh2)
        self.assertEqual(h3, sh3)
        self.assertIn(h1, shs)
        self.assertIn(h2, shs)
        self.assertIn(h3, shs)
        self.assertIn(sh1, hs)
        self.assertIn(sh2, hs)
        self.assertIn(sh3, hs)

    def testPort(self):
        portA = network.PortForward("test-ssh", 1, "tcp", "", "9866", "", "22")
        portB = network.PortForward("test-http", 1, "tcp", "", "9867", "", "80")
        portC = network.PortForward("test-https", 1, "tcp", "", "9868", "", "443")
        sportA = network.PortForward("test-ssh", 1, "tcp", "", "9866", "", "22")
        sportB = network.PortForward("test-http", 1, "tcp", "", "9867", "", "80")
        sportC = network.PortForward("test-https", 1, "tcp", "", "9868", "", "443")

        ports = [portA, portB, portC]
        sports = [sportA, sportB, sportC]

        self.assertTrue(portA == sportA)
        self.assertTrue(portB == sportB)
        self.assertTrue(portC == sportC)
        self.assertIn(portA, sports)
        self.assertIn(portB, sports)
        self.assertIn(portC, sports)
        self.assertIn(sportA, ports)
        self.assertIn(sportB, ports)
        self.assertIn(sportC, ports)

    def doImport(self):
        ovfFile = "/Users/bbeausej/dev/substance-engine/box.ovf"
        op = machine.inspectOVF(ovfFile) \
            .bind(defer(machine.makeImportParams, name=self.vmName)) \
            .bind(defer(machine.importOVF, ovfFile=ovfFile, name=self.vmName))
        self.assertIsInstance(op, OK)
        self.assertIsNotNone(op.getOK())

        type(self).vmUUID = op.getOK()

        op = machine.readMachineExists(self.vmUUID)
        self.assertIsInstance(op, OK)

        self.assertTrue(op.getOK())

    def doTestPF(self):
        portA = network.PortForward("test-ssh", 1, "tcp", "", "9866", "", "22")
        portB = network.PortForward("test-http", 1, "tcp", "", "9867", "", "80")
        portC = network.PortForward("test-https", 1, "tcp", "", "9868", "", "443")
        ports = [portA, portB, portC]

        op = network.addPortForwards(ports, self.vmUUID)  \
            .then(defer(network.readPortForwards, uuid=self.vmUUID))
        self.assertIsInstance(op, OK)

        setports = op.getOK()
        self.assertEqual(type(setports), type([]))

        for port in setports:
            self.assertIsInstance(port, network.PortForward)

        for port in ports:
            self.assertIn(port, setports)

        op = network.removePortForwards([portB], self.vmUUID) \
            .then(defer(network.readPortForwards, uuid=self.vmUUID))
        self.assertIsInstance(op, OK)

        self.assertIn(portA, op.getOK())
        self.assertNotIn(portB, op.getOK())
        self.assertIn(portC, op.getOK())

        op = network.clearPortForwards(self.vmUUID) \
            .then(defer(network.readPortForwards, uuid=self.vmUUID))

        self.assertIsInstance(op, OK)
        self.assertEqual(op.getOK(), [])

    def doTestSF(self):
        d1 = self.addTemporaryDir()
        f1 = machine.SharedFolder(os.path.basename(d1), d1)
        d2 = self.addTemporaryDir()
        f2 = machine.SharedFolder(os.path.basename(d2), d2)
        folders = [f1, f2]

        op = machine.addSharedFolders(folders, self.vmUUID) \
            .then(defer(machine.readSharedFolders, uuid=self.vmUUID))
        self.assertIsInstance(op, OK)
        self.assertEqual(len(op.getOK()), 2)
        self.assertEqual([x.hostPath for x in op.getOK()], [d1, d2])

        op = machine.clearSharedFolders(self.vmUUID) \
            .then(defer(machine.readSharedFolders, uuid=self.vmUUID))
        self.assertIsInstance(op, OK)
        self.assertEqual(len(op.getOK()), 0)

    def doStart(self):
        self.assertIsInstance(machine.start(self.vmUUID), OK)
        self.checkStateMatch("running")

    def doSuspend(self):
        self.assertIsInstance(machine.suspend(self.vmUUID), OK)
        self.checkStateMatch("saved")

    def doTerminate(self):
        self.assertIsInstance(machine.terminate(self.vmUUID), OK)
        self.checkStateMatch("poweroff")

    def doDelete(self):
        self.assertIsInstance(machine.delete(self.vmUUID), OK)
        op = machine.readMachineExists(self.vmUUID)
        self.assertIsInstance(op, OK)
        self.assertFalse(op.getOK())
        type(self).vmUUID = None

    def checkStateMatch(self, stateMatch):
        state = machine.readMachineState(self.vmUUID)
        self.assertIsInstance(state, OK)
        self.assertEqual(state.getOK(), stateMatch)

    def doReadGuestAdd(self):
        op = machine.readGuestAddVersion(self.vmUUID)
        self.assertIsInstance(op, OK)
        self.assertTrue(re.match(r'^[0-9\.]*$', op.getOK()))



import os

from substance.monads import OK


class BaseSyncher(object):
    def __init__(self, engine, keyfile):
        self.keyfile = keyfile
        self.engine = engine

    def ensureKeyPerms(self):
        try:
            # XXX Hack for perms on insecure file. Temporary until distribution is sorted out.
            os.chmod(self.keyfile, 0o600)
        except Exception as err:
            pass
        return OK(None)



import os
import logging
from collections import OrderedDict
from substance.logs import *
from substance.monads import *
from substance.config import Config
from substance.driver import Driver
from substance.constants import (EngineStates)
from substance.exceptions import (FileDoesNotExist)

from .vbox import (vboxManager)
from . import network
from . import machine
from .exceptions import *

from netaddr import (IPAddress, IPNetwork)
from functools import reduce

logger = logging.getLogger(__name__)


class VirtualBoxDriver(Driver):
    '''
    Substance VirtualBox driver class. Interface to virtual box manager.
    '''

    def __init__(self, core):
        super(self.__class__, self).__init__(core)
        self.config = Config(os.path.join(
            self.core.getBasePath(), "virtualbox.yml"))

    def assertSetup(self):
        return self.assertConfig().then(self.assertNetworking)

    # -- Configuration

    def getDefaultConfig(self):
        defaults = OrderedDict()
        defaults['network'] = "172.21.21.0/24"
        defaults['interface'] = None
        return defaults

    def makeDefaultConfig(self):
        self.logAdapter.info(
            "Generating default virtualbox config in %s" % self.config.getConfigFile())
        defaults = self.getDefaultConfig()
        for kkk, vvv in defaults.items():
            self.config.set(kkk, vvv)
        return self.config.saveConfig()

    def assertConfig(self):
        return self.config.loadConfigFile() \
            .catchError(FileDoesNotExist, lambda err: self.makeDefaultConfig())

    # ---- Networking setup

    def getNetworkInterface(self):
        return self.config.get('interface', None)

    def readNetworkConfig(self):
        netrange = IPNetwork(self.config.get('network'))
        netconfig = {
            'gateway': netrange[1].format(),
            'netmask': netrange.netmask.format(),
            'lowerIP': netrange[5].format(),
            'upperIP': netrange[-5].format()
        }

        return OK(netconfig)

    def assertNetworking(self):
        interface = self.getNetworkInterface()
        netconfig = self.readNetworkConfig()

        if interface:
            hoif = network.readHostOnlyInterface(interface) \
                .catchError(VirtualBoxObjectNotFound, lambda err: OK(None))
            dhcp = network.readDHCP(interface) \
                .catchError(VirtualBoxObjectNotFound, lambda err: OK(None))

            return Try.sequence((netconfig, hoif, dhcp)) \
                .bind(self.assertNetworkConfiguration)
        else:
            return netconfig.bind(self.provisionNetworking)

    def assertNetworkConfiguration(self, netparts):

        (netconfig, hoif, dhcp) = netparts

        if not hoif:
            return self.provisionNetworking(netconfig)
        elif hoif.v4ip.format() != netconfig['gateway']:
            self.logAdapter.warning(
                "VirtualBox interface \"%s\" is not properly configured. Creating a new host-only network." % hoif.name)
            return self.provisionNetworking(netconfig)
        elif dhcp is None:
            self.logAdapter.warning(
                "VirtualBox interface \"%s\" does not have DHCP enabled. Re-Establishing now." % hoif.name)
            return self.provisionDHCP(hoif.name, netconfig)

        return OK(hoif)

    def findInterface(self, netconfig, interfaces):
        matches = list(filter(lambda x: x.v4ip == netconfig['gateway'], interfaces))
        return OK(None) if len(matches) <= 0 else OK(matches[0])

    def provisionNetworking(self, netconfig):
        self.logAdapter.info(
            "Provisioning VirtualBox networking for substance")

        hoif = network.readHostOnlyInterfaces() \
            .bind(defer(self.findInterface, netconfig))
        if hoif.isFail():
            return hoif
        hoif = hoif.getOK()

        if not hoif:
            ifm = network.addHostOnlyInterface()
            if ifm.isFail():
                return ifm
            iface = ifm.getOK()
        else:
            iface = hoif.name

        return network.configureHostOnlyInterface(iface, ip=netconfig['gateway'], netmask=netconfig['netmask']) \
            .then(defer(self.provisionDHCP, interface=iface, netconfig=netconfig)) \
            .then(defer(self.saveInterface, iface=iface)) \
            .then(defer(network.readHostOnlyInterface, name=iface))

    def provisionDHCP(self, interface, netconfig):
        self.logAdapter.info(
            "Provisioning DHCP service for host only interface")
        network.removeDHCP(interface).catch(lambda x: OK(interface)) \
            .then(defer(network.addDHCP, interface, **netconfig))

    def saveInterface(self, iface):
        self.config.set('interface', iface)
        return self.config.saveConfig()

    # ---- Machine API

    def importMachine(self, name, ovfFile, engineProfile=None):
        return self.assertSetup() \
            .then(defer(machine.inspectOVF, ovfFile)) \
            .bind(defer(machine.makeImportParams, name=name, engineProfile=engineProfile)) \
            .bind(defer(machine.importOVF, ovfFile=ovfFile, name=name))

    def startMachine(self, uuid):
        '''
        Start the machine by driver identifier.
        '''
        state = machine.readMachineState(uuid)
        if state.isFail():
            return state

        if state.getOK() in [machine.MachineStates.PAUSED]:
            return machine.resume(uuid)
        else:
            return machine.start(uuid)

    def readMachineWaitForNetwork(self, uuid, timeout=5000):
        return machine.readWaitGuestProperty(uuid, "/VirtualBox/GuestInfo/Net", timeout)

    def readMachineNetworkInfo(self, uuid):
        def format(res):
            info = OrderedDict()
            info['sshPort'] = res[0].hostPort if res[0] else None
            info['sshIP'] = '127.0.0.1'
            info['privateIP'] = res[1]
            info['publicIP'] = res[2]
            return info

        return Try.sequence([
            network.readPortForward(uuid, name="substance-ssh"),
            machine.readGuestProperty(
                uuid, "/VirtualBox/GuestInfo/Net/0/V4/IP"),
            machine.readGuestProperty(
                uuid, "/VirtualBox/GuestInfo/Net/1/V4/IP")
        ]).map(format)

    def configureMachine(self, uuid, engine):

        def engineFolderToVBoxFolder(acc, x):
            if x.mode == 'sharedfolder':
                acc.append(machine.SharedFolder(
                    name=x.name, hostPath=x.hostPath))
            return acc

        folders = reduce(engineFolderToVBoxFolder,
                         engine.getEngineFolders(), [])
        desiredPort = engine.getSSHPort()

        return self.assertSetup() \
            .then(defer(self.resolvePortConflict, uuid=uuid, desiredPort=desiredPort)) \
            .then(defer(self.configureMachineAdapters, uuid=uuid)) \
            .then(defer(self.configureMachineFolders, uuid=uuid, folders=folders)) \
            .then(defer(self.configureMachineProfile, uuid=uuid, engine=engine))

    def configureMachineProfile(self, uuid, engine):
        self.logAdapter.info("Configure machine profile")
        profile = engine.getProfile()
        return machine.configureProfile(uuid, profile.cpus, profile.memory)

    def configureMachineFolders(self, folders, uuid):
        self.logAdapter.info("Configure machine shared folders")
        return machine.clearSharedFolders(uuid) \
            .then(defer(machine.addSharedFolders, folders=folders, uuid=uuid))

    def resolvePortConflict(self, uuid, desiredPort):
        self.logAdapter.info("Checking for port conflicts")
        return network.readAllPortForwards(ignoreUUIDs=[uuid]) \
            .bind(defer(self.determinePort, desiredPort=desiredPort)) \
            .bind(defer(self.configurePort, uuid=uuid))

    def determinePort(self, usedPorts, desiredPort):
        basePort = 4500
        self.logAdapter.debug("Base port: %s" % basePort)
        self.logAdapter.debug("Desired port: %s" % desiredPort)
        unavailable = [x.hostPort for x in usedPorts]
        port = desiredPort if desiredPort >= basePort else basePort
        while port in unavailable or port < basePort:
            self.logAdapter.debug("Port %s is in use." % port)
            port += 1
        self.logAdapter.info("Determined SSH port as %s" % port)
        return OK(port)

    def configurePort(self, port, uuid):
        self.logAdapter.debug("Configure substance-ssh port on port %s" % port)
        pf = network.PortForward(
            "substance-ssh", 1, "tcp", None, port, None, 22)
        return network.removePortForwards([pf], uuid) \
            .catch(lambda err: OK(None)) \
            .then(defer(network.addPortForwards, [pf], uuid))

    def configureMachineAdapters(self, uuid):
        interface = self.getNetworkInterface()
        adapters = [
            machine.AdapterSettings(
                machine.AdapterTypes.NAT, 'default', None, False),
            machine.AdapterSettings(
                machine.AdapterTypes.HOSTONLY, interface, None, False),
            machine.AdapterSettings(
                machine.AdapterTypes.NONE, None, None, False),
            machine.AdapterSettings(
                machine.AdapterTypes.NONE, None, None, False)
        ]
        return Try.sequence([machine.configureAdapter(uuid, i + 1, x) for i, x in enumerate(adapters)])

    def suspendMachine(self, uuid):
        '''
        Suspend the machine.
        '''
        return machine.suspend(uuid)

    def haltMachine(self, uuid):
        '''
        Halt the machine.
        '''
        return machine.halt(uuid)

    def terminateMachine(self, uuid):
        '''
        Terminate the machine forcefully.
        '''
        return machine.terminate(uuid)

    def deleteMachine(self, uuid):
        '''
        Delete the machine by driver identifier.
        '''
        return machine.delete(uuid)

    def exportMachine(self, uuid):
        # XXX To be implemented
        pass

    def getMachines(self):
        '''
        Retrieve the list of machines and their driver identifiers.
        '''
        return machine.readMachines()

    # -- Parse results from Virtual Box

    def getMachineID(self, name):
        '''
        Retrieve the driver specific machine ID for a machine name.
        '''
        return machine.findMachineID(name)

    def exists(self, uuid):
        '''
        Check in the driver that the specified identifier exists.
        '''
        return machine.readMachineExists(uuid)

    def isRunning(self, uuid):
        if self.getMachineState(uuid) is EngineStates.RUNNING:
            return True

    def isStopped(self, uuid):
        if self.getMachineState(uuid) is not EngineStates.RUNNING:
            return True

    def isSuspended(self, uuid):
        if self.getMachineState(uuid) is EngineStates.SUSPENDED:
            return True

    def getMachineState(self, uuid):
        '''
        Retrieve the Substance machine state for this driver id
        '''
        return machine.readMachineState(uuid) \
            .bind(self.machineStateToEngineState)

    def machineStateToEngineState(self, vboxState):
        '''
        Resolve a vbox machine state to a substance engine state.
        '''

        mapping = {
            machine.MachineStates.POWEROFF: EngineStates.STOPPED,
            machine.MachineStates.SAVED: EngineStates.SUSPENDED,
            machine.MachineStates.PAUSED: EngineStates.SUSPENDED,
            machine.MachineStates.ABORTED: EngineStates.STOPPED,
            machine.MachineStates.STUCK: EngineStates.STOPPED,
            machine.MachineStates.RESTORING: EngineStates.STOPPED,
            machine.MachineStates.SNAPSHOTTING: EngineStates.STOPPED,
            machine.MachineStates.SETTING_UP: EngineStates.STOPPED,
            machine.MachineStates.ONLINE_SNAPSHOTTING: EngineStates.STOPPED,
            machine.MachineStates.RESTORING_SNAPSHOT: EngineStates.STOPPED,
            machine.MachineStates.DELETING_SNAPSHOT: EngineStates.STOPPED,
            machine.MachineStates.LIVE_SNAPSHOTTING: EngineStates.RUNNING,
            machine.MachineStates.RUNNING: EngineStates.RUNNING,
            machine.MachineStates.STARTING: EngineStates.RUNNING,
            machine.MachineStates.STOPPING: EngineStates.RUNNING,
            machine.MachineStates.SAVING: EngineStates.RUNNING,
            machine.MachineStates.UNKNOWN: EngineStates.UNKNOWN,
            machine.MachineStates.INACCESSIBLE: EngineStates.INEXISTENT,
            machine.MachineStates.INEXISTENT: EngineStates.INEXISTENT
        }
        state = mapping.get(vboxState, EngineStates.UNKNOWN)
        ddebug("Machine state: %s : %s", vboxState, state)
        return OK(state)



import logging
import functools
from functools import (wraps)
from functools import partial
from functools import reduce


class Monad(object):

    def bind(self, mf):
        raise NotImplementedError

    def map(self, mf):
        raise NotImplementedError

    def __rshift__(self, f):
        return self.bind(f)

    def __ilshift__(self, f):
        return self.bind(f)


class Maybe(Monad):

    @staticmethod
    def of(value=None):
        if(value is None):
            return Nothing()
        else:
            return Just(value)

    def get(self):
        raise NotImplementedError

    def isNothing(self):
        if isinstance(self, Nothing):
            return True

    def bind(self, mf):
        if self.isNothing():
            return self
        return mf(self.get())

    def map(self, f):
        if self.isNothing():
            return self
        return Just(f(self.get()))

    def getOrElse(self, default):
        return default if self.isNothing() else self.get()


class Just(Maybe):
    value = None

    def __init__(self, value):
        self.value = value

    def get(self):
        return self.value

    def __repr__(self):
        return "Just(%r)" % self.value

    def __eq__(self, other):
        if isinstance(other, Just):
            return self.value == other.value
        return False


class Nothing(Maybe):
    def get(self):
        raise NoValue

    def __repr__(self):
        return "Nothing"

    def __eq__(self, other):
        return isinstance(other, Nothing)


class Either(Monad):

    def isLeft(self):
        return isinstance(self, Left)

    def isRight(self):
        return isinstance(self, Right)

    def getLeft(self):
        raise NotImplementedError

    def getRight(self):
        raise NotImplementedError

    def bind(self, mf):
        if self.isLeft():
            return self
        return mf(self.getRight())

    def map(self, mf):
        if self.isLeft():
            return self
        return Right(mf(self.getRight()))


class Left(Either):
    value = None

    def __init__(self, value):
        self.value = value

    def getLeft(self):
        return self.value

    def __repr__(self):
        return "Left(%r)" % self.value

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False


class Right(Either):
    value = None

    def __init__(self, value):
        self.value = value

    def getRight(self):
        return self.value

    def __repr__(self):
        return "Right(%r)" % self.value

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False


class Try(Monad):

    @staticmethod
    def of(value):
        return OK(value)

    def isFail(self):
        return isinstance(self, Fail)

    def isOK(self):
        return isinstance(self, OK)

    def getError(self):
        raise NotImplementedError

    def getOK(self):
        raise NotImplementedError

    def getOrElse(self, default):
        return default if self.isFail() else self.getOK()

    def bind(self, mf, *args, **kwargs):
        if self.isFail():
            return self
        return mf(self.getOK(), *args, **kwargs)

    def map(self, f):
        if self.isFail():
            return self
        return OK(f(self.getOK()))

    def thenIfTrue(self, f):
        return self.thenIfBool(f, lambda: OK(False))

    def thenIfFalse(self, f):
        return self.thenIfBool(lambda: OK(True), f)

    def thenIfNone(self, f):
        if self.isFail():
            return self
        if self.getOK() is None:
            return self.then(f)
        else:
            return self

    def thenIfBool(self, fTrue, fFalse):
        if self.isFail():
            return self
        if self.getOK() is True:
            return self.then(fTrue)
        else:
            return self.then(fFalse)

    def then(self, okF=None, failF=None):
        if self.isFail():
            if failF:
                r = failF(self.getError())
                if isinstance(r, Try):
                    return r
                return self
            else:
                return self
        else:
            if okF:
                r = okF()
                if isinstance(r, Try):
                    return r
                return OK(r)
            else:
                return self

    def catch(self, f):
        return self.then(None, f)

    def catchError(self, err, f):
        def catcher(x):
            if isinstance(x, err):
                return f(self.getError())
        return self.then(None, catcher)

    def bindIfTrue(self, f):
        return self.bindIf(f, lambda x: OK(False))

    def bindIfFalse(self, f):
        return self.bindIf(lambda x: OK(True), f)

    def bindIf(self, fTrue, fFalse):
        if self.isFail():
            return self
        if self.getOK() is True:
            return self.bind(fTrue)
        else:
            return self.bind(fFalse)

    def mapM(self, mf):
        def mapper(xs):
            return mapM(self, mf, xs)
        return self.bind(mapper)

    def mapM_(self, mf):
        def mapper(xs):
            return mapM_(self, mf, xs)
        return self.bind(mapper)

    @staticmethod
    def compose(*funcs):
        def func(x):
            return fold(lambda acc, f: acc.bind(f), funcs, Try.of(x))
        return func

    @staticmethod
    def sequence(monads):
        ''' Fold a list of monads into a monad containing the list of values '''
        return reduce(lambda acc, mv: unshiftM(Try, acc, mv), reversed(monads), Try.of([]))

    @staticmethod
    def attemptDeferred(f, expect=Exception):
        def tryer(*args, **kwargs):
            try:
                return OK(f(*args, **kwargs))
            except expect as e:
                return Fail(e)
        return tryer

    @staticmethod
    def attempt(f, *args, **kwargs):
        return Try.attemptDeferred(f)(*args, **kwargs)

    @staticmethod
    def raiseError(self, err=Exception):
        raise(err)

    def join(self):
        return self if self.isFail() else self.getOK()


class OK(Try):
    value = None

    def __init__(self, value):
        self.value = value

    def getOK(self):
        return self.value

    def __repr__(self):
        return "OK(%r)" % self.value

    def __bool__(self):
        return True

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False


class Fail(Try):
    value = None

    def __init__(self, value):
        self.value = value

    def getError(self):
        return self.value

    def __bool__(self):
        return False

    def __repr__(self):
        return "Fail(%r)" % self.value

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False


def fold(f, xs, init=None):
    return functools.reduce(f, xs, init)


def compose(*funcs):
    def func(x):
        return fold(lambda acc, f: f(acc), reversed(funcs), x)
    return func


def unshiftM(monad, monads, mv):
    ''' Prepend a monadic value to the list value of a monad '''
    return monads.bind(lambda xs: mv.bind(lambda x: monad.of([x] + xs)))


def sequence(monad, monads):
    ''' Fold a list of monads into a monad containing the list of values '''
    return reduce(lambda acc, mv: unshiftM(monad, acc, mv), reversed(monads), monad.of([]))


def mapM(monad, mf, xs):
    ''' Map a monadic function over a list of values, lifting them into monadic context and convert the results to a monad containing a list of the values. '''
    mapped = list(map(mf, xs))
    sequenced = sequence(monad, mapped)
    return sequenced


def mapM_(monad, mf, xs):
    ''' Map a monadic  action to a structure, evaluate from left to right and ignore results. '''
    list(map(mf, xs))
    return monad.of(xs)


def defer(f, *args, **kwargs):
    fargs = args
    fkwargs = kwargs
    return partial(f, *fargs, **fkwargs)


def failWith(exception):
    return (lambda *x: Fail(exception))


def chainSelf(self, *_, **__):
    return self


def flatten(xs):
    return [x for sl in xs for x in sl]



import os
import subprocess
from substance.monads import *
from substance.constants import *
from substance.logs import dinfo
from substance import Shell
from substance.subenv import SubenvSpec
from substance.exceptions import (InvalidEnvError, InvalidOptionError)
from substance.utils import makeSymlink, readSymlink

logger = logging.getLogger(__name__)


class SubenvAPI(object):
    '''
      Subenv API
    '''

    def __init__(self, basePath="/substance"):
        self.basePath = basePath
        self.envsPath = os.path.join(self.basePath, "envs")
        self.assumeYes = False
        self.struct = {'dirs': [], 'files': []}

    def setAssumeYes(self, b):
        if b:
            self.assumeYes = True
        else:
            self.assumeYes = False

    def initialize(self):
        return self.assertPaths()

    def assertPaths(self):
        return OK([self.basePath, self.envsPath]).mapM(Shell.makeDirectory)

    def init(self, path, env={}):
        logger.info("Initializing subenv from: %s" % path)
        return SubenvSpec.fromSpecPath(path, env) \
            .bind(self._applyEnv)

    def exists(self, name):
        if os.path.isdir(os.path.join(self.envsPath, name)):
            return OK(True)
        return OK(False)

    def vars(self, envName=None, vars=None):
        if not envName:
            envSpec = self._getCurrentEnv()
            if not envSpec:
                return Fail(InvalidEnvError("No env is currently active."))
            envName = envSpec.name

        def varfilter(o):
            if vars is not None:
                return {k: v for k, v in o.items() if k in vars}
            else:
                return o

        return OK(envName) \
            .bind(self._loadEnvSpec) \
            .map(lambda e: e.getEnvVars()) \
            .map(varfilter)

    def delete(self, name):
        envPath = os.path.normpath(os.path.join(self.envsPath, name))
        if not os.path.isdir(envPath):
            return Fail(InvalidOptionError("Environment '%s' does not exist."))
        return Shell.nukeDirectory(envPath)

    def use(self, name):
        envPath = os.path.normpath(os.path.join(self.envsPath, name))
        if not os.path.isdir(envPath):
            return Fail(InvalidOptionError("Environment '%s' does not exist."))

        envSpec = SubenvSpec.fromEnvPath(envPath)
        current = os.path.join(self.basePath, "current")

        return Try.attempt(makeSymlink, envSpec.envPath, current, True) \
            .then(dinfo("Current substance environment now: '%s'" % envSpec.name))

    def current(self):
        return OK(self._getCurrentEnv())

    def run(self, args, envName=None):
        if not envName:
            envSpec = self._getCurrentEnv()
            if not envSpec:
                return Fail(InvalidEnvError("No env is currently active."))
            envName = envSpec.name

        cmd = subprocess.list2cmdline(args)

        return OK(envName) \
            .bind(self._loadEnvSpec) \
            .map(lambda x: x.envPath) \
            .bind(lambda p: Shell.call(cmd, cwd=p, shell=True))

    def ls(self):
        envs = []
        current = self._getCurrentEnv()
        for f in os.listdir(self.envsPath):
            path = os.path.join(self.envsPath, f)
            if os.path.isdir(path):
                env = SubenvSpec.fromEnvPath(path)
                if current and env.envPath == current.envPath:
                    env.current = True
                envs.append(env)
        return OK(envs)

    def _loadEnvSpec(self, name):
        envPath = os.path.normpath(os.path.join(self.envsPath, name))
        if not os.path.isdir(envPath):
            return Fail(InvalidOptionError("Environment '%s' does not exist." % name))
        return OK(SubenvSpec.fromEnvPath(envPath))

    def _getCurrentEnv(self):
        try:
            current = readSymlink(os.path.join(self.basePath, "current"))
            return SubenvSpec.fromEnvPath(current)
        except Exception as err:
            return None

    def _applyEnv(self, envSpec):
        envPath = os.path.join(self.envsPath, envSpec.name)
        logger.info("Applying environment to: %s" % envPath)
        return envSpec.applyTo(envPath)



from substance.monads import *
from substance.logs import *
from substance.utils import mergeDictOverwrite
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)


class Aliases(Command):

    def getUsage(self):
        return "substance aliases [opts]"

    def getHelpTitle(self):
        return "Execute a command within a container"

    def getShellOptions(self, optparser):
        optparser.add_option("-p", "--prefix", dest="prefix",
                             help="Prefix of aliases", default="")
        optparser.add_option("-s", "--suffix", dest="suffix",
                             help="Suffix of aliases", default="")
        return optparser

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.envLoadCurrent) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.loadSubenvConfigFile) \
            .bind(self.exportAliases) \
            .catch(self.exitError)

    def exportAliases(self, engine):
        aliases = engine.config.get('aliases')
        if engine.subenvConfig.get('aliases'):
            aliases = mergeDictOverwrite(engine.config.get('aliases'),
                                         engine.subenvConfig.get('aliases'))
        if aliases:
            prefix = self.getOption('prefix')
            suffix = self.getOption('suffix')
            exports = ["alias %s%s%s=\"substance -e %s %s\"" %
                       (prefix, alias, suffix, engine.name, alias) for alias in list(aliases.keys())]
            print("\n".join(exports))
            return OK(True)
        else:
            return Fail("Engine %s has no aliases." % engine.name)



import logging
from substance.logs import *
from substance.monads import *
from substance import (Command, Engine, Shell)
from substance.exceptions import (SubstanceError, EngineNotProvisioned)


class Delete(Command):

    def getUsage(self):
        return "substance engine delete [ENGINE NAME]"

    def getHelpTitle(self):
        return "Permanently delete an engine and it's associated machine."

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        name = self.getInputName()
        self.ask("You are about to delete engine \"%s\"" % name) \
            .then(defer(self.core.loadEngine, name)) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.deprovision) \
            .catchError(EngineNotProvisioned, lambda e: OK(None)) \
            .then(defer(self.core.removeEngine, name)) \
            .then(dinfo("Engine \"%s\" has been deleted.", name)) \
            .catch(self.exitError)



from substance.monads import *
from substance.logs import *
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)


class Env(Command):

    def getUsage(self):
        return "substance engine env [ENGINE NAME]"

    def getHelpTitle(self):
        return "Print the shell variables to set up the local docker client environment"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):

        name = self.getInputName()

        self.core.loadEngine(name) \
            .bind(Engine.loadConfigFile) \
            .bind(self.outputDockerEnv) \
            .catch(self.exitError)

    def outputDockerEnv(self, engine):
        env = engine.getDockerEnv()
        for k, v in env.items():
            print(("export %s=\"%s\"" % (k, v)))
        return OK(None)



from substance.monads import *
from substance.logs import *
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)
import shlex


class Exec(Command):

    def getUsage(self):
        return "substance exec [opts] [CONTAINER] [CMD...]"

    def getHelpTitle(self):
        return "Execute a command within a container"

    def getShellOptions(self, optparser):
        optparser.add_option("-u", "--user", dest="user",
                             help="User to connect", default=None)
        optparser.add_option("-d", "--cwd", dest="cwd",
                             help="Change to this directory upon connect", default=None)
        return optparser

    # Disable intersped args
    def execute(self, args=None):
        self.input = args
        (self.options, self.args) = self.parseShellInput(False)

        if len(self.args) == 0:
            return self.exitError("Please provide a container name to exec on.")

        self.container = self.args[0]
        self.main()

    def main(self):
        container = self.getInputContainer()
        if not container:
            return self.exitError("Please provide a container name to exec on.")

        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envExec, container=container, user=self.getOption('user'), cwd=self.getOption('cwd'), cmd=self.args[1:]) \
            .catch(self.exitError)

    def getInputContainer(self):
        if len(self.args) <= 0:
            return None
        return self.args[0]




from platform import (system, uname)
from substance.utils import Memoized

@Memoized
def isCygwin():
    return system().startswith("CYGWIN")

@Memoized
def isWSL():
    return "microsoft" in uname()[3].lower()

@Memoized
def isWithinWindowsSubsystem():
    return (isCygwin() or isWSL())


import sys
import platform
import os
import re
from substance.shell import (Shell)
from subprocess import check_output
from substance.utils import Memoized
from pathlib import (PureWindowsPath, PurePosixPath)

from substance.platform import (isWSL, isCygwin)


@Memoized
def inner(path):
    if isCygwin():
        path = check_output(["cygpath", "-u", path]).decode().strip()
    if isWSL() and PureWindowsPath(path).drive:
        path = check_output(["wslpath", "-u", path]).decode().strip()
    return path


@Memoized
def outer(posixPath):
    if PureWindowsPath(posixPath).drive:
        path = posixPath
    elif isCygwin():
        path = check_output(["cygpath", "-w", posixPath]).decode().strip()
    elif isWSL():
        path = check_output(["wslpath", "-w", posixPath]).decode().strip()
    else:
        path = posixPath
    return path


@Memoized
def getHomeDirectory(subpath=None):
    if isWSL():
        homePath = getWindowsHomeDirectory()
    else:
        homePath = getUnixHomeDirectory()

    path = homePath.joinpath(subpath) if subpath else homePath
    return inner(str(path))


@Memoized
def getUnixHomeDirectory():
    return PurePosixPath(os.path.expanduser('~'))


@Memoized
def getWindowsHomeDirectory():
    winPath = PureWindowsPath(check_output(["cmd.exe", "/C", 'echo %USERPROFILE%']).decode().strip())
    return PurePosixPath(winPath.as_posix())


def pathComponents(path):
    folders = []
    path = os.path.normpath(path)
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)
            break

    folders.reverse()
    return folders


def expandLocalPath(path, basePath=None):
    path = os.path.expanduser(path)
    if not os.path.isabs(path) and basePath:
        path = os.path.normpath(path)
        path = os.path.join(basePath, path)
    return os.path.abspath(os.path.realpath(path))



import unittest
import os
import tempfile
import tests
from substance.core import Core
from substance.engine import (EngineProfile, Engine)
from substance.shell import Shell
from substance.monads import *


class TestCore(tests.TestBase):

    def setUp(self):
        self.basePath = self.addTemporaryDir()
        self.projectsPath = self.addTemporaryDir()

    def tearDown(self):
        if self.basePath:
            Shell.nukeDirectory(self.basePath).catch(TestCore.raiser)
        if self.projectsPath:
            Shell.nukeDirectory(self.projectsPath).catch(TestCore.raiser)

    def setUpDefaultCore(self):
        self.core = Core(basePath=self.basePath)
        self.assertIsInstance(self.core.initialize(), OK)
        return self.core

    def testInitialize(self):
        core = Core(basePath=self.basePath)
        self.assertIsInstance(core.initialize(), OK)
        self.assertEqual(core.getBasePath(), self.basePath)
        self.assertTrue(os.path.exists(self.basePath))
        self.assertTrue(os.path.exists(core.getBasePath()))
        self.assertTrue(os.path.exists(core.getEnginesPath()))
        self.assertTrue(os.path.exists(os.path.join(self.basePath, 'substance.yml')))

        core = Core(basePath=self.basePath, configFile="foobar.yml")
        self.assertIsInstance(core.initialize(), OK)
        self.assertTrue(os.path.exists(self.basePath))
        self.assertTrue(os.path.exists(os.path.join(self.basePath, 'foobar.yml')))

    def testDefaultConfig(self):
        self.setUpDefaultCore()
        self.assertIsNotNone(self.core.config)
        self.assertEqual(self.core.config.get('assumeYes', None), False)
        self.assertEqual(self.core.config.get('basePath', None), self.basePath)

    def testConfig(self):
        self.setUpDefaultCore()
        self.assertEqual(self.core.config.set("foo", "bar"), "bar")
        self.assertEqual(self.core.config.get("foo", None), "bar")
        self.assertIsInstance(self.core.config.saveConfig(), OK)
        self.assertIsInstance(self.core.config.loadConfigFile(), OK)
        self.assertEqual(self.core.config.get("foo", None), "bar")

    def coreTestEngines(self):
        self.setUpDefaultCore()
        self.assertEqual(self.core.config.set("assumeYes", True), True)

        engines = self.core.getEngines()
        self.assertIsInstance(engines, OK)
        self.assertEqual(engines.getOK(), [])

    def testCreateEngine(self):
        self.setUpDefaultCore()
        createOp = self.core.createEngine("coreTestEngine", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp, OK)

        engine = self.core.loadEngine("coreTestEngine")
        self.assertIsInstance(engine, OK)

        self.assertEqual(os.path.join(self.core.getEnginesPath(), "coreTestEngine"), engine.getOK().getEnginePath())
        self.assertTrue(engine.getOK().getEnginePath(), "coreTestEngine")

        engine = engine.bind(Engine.loadConfigFile)
        self.assertIsInstance(engine, OK)

        engine = engine.getOK()

        self.assertEqual(engine.config.get('projectsPath'), self.projectsPath)
        self.assertEqual(engine.config.get('profile').get('cpus'), 4)
        self.assertEqual(engine.config.get('profile').get('memory'), 512)

    def testGetEngines(self):
        self.setUpDefaultCore()
        createOp = self.core.createEngine("coreTestEngine", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp, OK)
        createOp2 = self.core.createEngine("coreTestEngine2", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp2, OK)
        createOp3 = self.core.createEngine("coreTestEngine3", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp3, OK)

        engines = self.core.getEngines()
        self.assertIsInstance(engines, OK)
        self.assertEqual(len(engines.getOK()), 3)
        self.assertEqual(engines.getOK(), ['coreTestEngine', 'coreTestEngine2', 'coreTestEngine3'])

    def testRemoveEngine(self):
        self.setUpDefaultCore()
        createOp = self.core.createEngine("coreTestEngine", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp, OK)
        createOp2 = self.core.createEngine("coreTestEngine2", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp2, OK)
        createOp3 = self.core.createEngine("coreTestEngine3", config={"projectsPath": self.projectsPath}, profile=EngineProfile(cpus=4, memory=512))
        self.assertIsInstance(createOp3, OK)

        engines = self.core.getEngines()
        self.assertIsInstance(engines, OK)

        self.core.removeEngine("coreTestEngine2")

        self.assertTrue(not os.path.exists(os.path.join(self.core.getEnginesPath(), "coreTestEngine2")))

        engines = self.core.getEngines()
        self.assertIsInstance(engines, OK)
        self.assertEqual(len(engines.getOK()), 2)
        self.assertEqual(engines.getOK(), ['coreTestEngine', 'coreTestEngine3'])

    def testKeys(self):
        self.setUpDefaultCore()
        k = self.core.getInsecureKeyFile()
        self.assertTrue(os.path.exists(k))

        k = self.core.getInsecurePubKeyFile()
        self.assertTrue(os.path.exists(k))




import sys
import logging
from collections import OrderedDict
from substance import (SubProgram, Core)


class Box(SubProgram):

    def __init__(self):
        super(Box, self).__init__()

    def setupCommands(self):
        self.addCommand('ls', 'substance.command.box.ls')
        self.addCommand('pull', 'substance.command.box.pull')
        self.addCommand('delete', 'substance.command.box.delete')
        return self

    def getShellOptions(self, optparser):
        return optparser

    def getUsage(self):
        return "substance box [options] COMMAND [command-options]"

    def getHelpTitle(self):
        return "Substance box management"

    def initCommand(self, command):
        command.core = self.core
        return command



import os
import logging
import shlex

from substance.logs import *
from substance.exceptions import (InvalidOptionError)
from substance import Command
from substance.subenv import (SubenvAPI)

logger = logging.getLogger(__name__)


class Current(Command):

    def getUsage(self):
        return "subenv current"

    def getHelpTitle(self):
        return "Print the current subenv on stdout"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        return self.api.current() \
            .thenIfNone(defer(self.exitError, "No current subenv is active.")) \
            .catch(self.exitError) \
            .bind(self.printEnv)

    def printEnv(self, envSpec):
        logger.info(envSpec.name)



import os
import logging
import shlex

from substance.logs import *
from substance.exceptions import (InvalidOptionError)
from substance import Command
from substance.subenv import (SubenvAPI)


class Use(Command):

    def getUsage(self):
        return "subenv use [ENV NAME]"

    def getHelpTitle(self):
        return "Set a project environment as current"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        name = self.readInputName().catch(self.exitError).getOK()
        return self.api.exists(name) \
            .thenIfFalse(defer(self.exitError, "Environment '%s' does not exist." % name)) \
            .then(defer(self.api.use, name))  \
            .catch(self.exitError)

    def readInputName(self):
        if len(self.args) <= 0:
            return Fail(InvalidOptionError("Please specify the name of a subenv."))
        name = os.path.normpath(self.args[0])
        return OK(name)



# -*- coding: utf-8 -*-
# $Id$

import traceback
from substance.logs import *
from substance.monads import *
from substance import (Engine, Command, Shell)
from substance.exceptions import (SubstanceError)


class Deprovision(Command):
    def getUsage(self):
        return "substance engine deprovision [ENGINE NAME]"

    def getHelpTitle(self):
        return "Destroy an engine's assocaited virtual machine"

    def main(self):
        name = self.getInputName()
        self.ask("You are about to deprovision engine \"%s\"" % name) \
            .then(defer(self.core.loadEngine, name)) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.deprovision) \
            .catch(self.exitError)



import sys
import logging
from importlib import import_module
from substance import (Program, Core)
from substance.subenv import SubenvAPI

import substance.logs


class SubenvCLI(Program):
    """Subenv CLI command"""

    def __init__(self):
        super(SubenvCLI, self).__init__()
        self.name = 'subenv'

    def getShellOptions(self, optparser):
        optparser.add_option("-d", '--debug',  dest="debug",
                             help="Activate debugging output", default=False, action="store_true")
        optparser.add_option('-y', "--yes", dest="assumeYes",
                             help="Assume yes when prompted", default=False, action="store_true")
        optparser.add_option('-b', '--base', type="str", dest="base",
                             help="Path to the base", default="/substance")
        return optparser

    def getUsage(self):
        return "subenv [options] COMMAND [command-options]"

    def getHelpTitle(self):
        return "Initialize a substance project environment"

    def setupCommands(self):
        self.addCommand('ls', 'substance.subenv.command.ls')
        self.addCommand('init', 'substance.subenv.command.init')
        self.addCommand('delete', 'substance.subenv.command.delete')
        self.addCommand('use', 'substance.subenv.command.use')
        self.addCommand('current', 'substance.subenv.command.current')
        self.addCommand('run', 'substance.subenv.command.run')
        self.addCommand('vars', 'substance.subenv.command.vars')

    def initCommand(self, command):
        api = SubenvAPI(self.getOption('base'))
        command.api = api
        return command


def cli():
    args = sys.argv[1:]
    prog = SubenvCLI()
    prog.execute(args)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate
from substance.constants import (EngineStates)
from substance.exceptions import (EngineNotRunning)

logger = logging.getLogger(__name__)


class Cleanup(Command):

    def getUsage(self):
        return "substance cleanup [options]"

    def getHelpTitle(self):
        return "Cleanup unused images and containers"

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.envCleanup) \
            .catch(self.exitError)



import os
import logging
import shlex

from substance.logs import *
from substance.exceptions import (InvalidOptionError)
from substance import Command
from substance.subenv import (SubenvAPI)
from substance.subenv.constants import (SPECDIR)

logger = logging.getLogger(__name__)


class Init(Command):

    def getUsage(self):
        return "subenv init [PATH TO PROJECT]"

    def getHelpTitle(self):
        return "Initialize or apply a project environment"

    def getShellOptions(self, optparser):
        optparser.add_option("-D", "--define", dest="define", default=[],
                             action='append', help="Define a variable for the environment")
        return optparser

    def main(self):
        params = [self.readInputPath(), self.readInputEnv()]
        return Try.sequence(params) \
            .bind(lambda l: self.api.init(*l)) \
            .catch(self.exitError) \
            .bind(lambda e: logger.info("Environment '%s' initialized." % e.name))

    def readInputEnv(self):
        env = {}
        envopts = self.options.define
        for opt in envopts:
            if '=' not in opt:
                return Fail(InvalidOptionError("%s is not a valid define. Use VAR=val form." % opt))
            (var, val) = ''.join(shlex.split(opt)).split('=')
            env[var] = val
        return OK(env)

    def readInputPath(self):
        if len(self.args) <= 0:
            return Fail(InvalidOptionError("Please specify a path to a '%s' folder." % SPECDIR))
        path = os.path.abspath(os.path.normpath(self.args[0]))
        return OK(path)



import os
import logging
import shlex
import sys

from substance.logs import *
from substance.exceptions import (InvalidOptionError)
from substance import Command
from substance.subenv import (SubenvAPI)

logger = logging.getLogger(__name__)


class Run(Command):

    def getUsage(self):
        return "subenv run"

    def getHelpTitle(self):
        return "Run a command in the current subenv directory"

    def getShellOptions(self, optparser):
        return optparser

    def parseShellInput(self, interspersed=True):
        """Disable parsing for this run command"""
        usage = self.getUsage()
        parser = self.getParser(interspersed)
        (opts, args) = parser.parse_args([])
        return (opts, args)

    def main(self):
        return self.api.run(args=self.input) \
            .catch(self.exitError)



import re
from collections import OrderedDict
from substance.monads import *
from substance.logs import *
from substance.constants import Constants
from .exceptions import *
from .vbox import (vboxManager, _vboxLineEnding)
from . import machine
from netaddr import (IPAddress, IPNetwork)

# -- Structs


class PortForward(object):
    def __init__(self, name, nic, proto, hostIP, hostPort, guestIP, guestPort):
        self.name = name
        self.nic = nic
        self.proto = proto
        self.hostIP = hostIP if hostIP else None
        self.hostPort = int(hostPort)
        self.guestIP = guestIP if guestIP else None
        self.guestPort = int(guestPort)

    def getCreateArg(self):

        def prune(some_dict):
            return { k: ('' if v is None else v) for k, v in some_dict.items() }

        return "--natpf{nic} \"{name}\",{proto},{hostIP},{hostPort},{guestIP},{guestPort}".format_map(prune(self.__dict__))

    def getDeleteArg(self):
        return "--natpf%(nic)s delete \"%(name)s\"" % self.__dict__

    def __repr__(self):
        return "PortForward(%(nic)s, %(name)s, %(proto)s host(%(hostIP)s:%(hostPort)s) -> guest(%(guestIP)s:%(guestPort)s))" % self.__dict__

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__repr__() == other.__repr__()
        return False


class DHCP(object):
    def __init__(self, interface, networkName, gateway, netmask, lowerIP, upperIP, enabled):
        self.interface = interface
        self.networkName = networkName
        self.gateway = gateway
        self.netmask = netmask
        self.lowerIP = lowerIP
        self.upperIP = upperIP
        self.enabled = True if enabled is True or enabled == "Yes" else False

        self.network = IPNetwork(gateway + "/" + netmask)

    def __repr__(self):
        rep = "DHCP(%(interface)s gateway: %(gateway)s netmask %(netmask)s (%(lowerIP)s to %(upperIP)s))" % self.__dict__
        rep += " enabled" if self.enabled else ""
        return rep

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__repr__() == other.__repr__()
        return False


class HostOnlyInterface(object):
    def __init__(self, name, mac=None, v4ip=None, v4mask=None, v6ip=None, v6prefix=None, status=None, netName=None):
        self.name = name
        self.mac = mac
        self.v4ip = v4ip
        self.v4mask = v4mask
        self.v6ip = v6ip
        self.v6prefix = v6prefix
        self.status = status
        self.netName = netName

        self.ip = IPAddress(v4ip)
        self.network = IPNetwork(v4ip + "/" + v4mask)

    def __repr__(self):
        return "HostOnlyInterface(%(name)s, %(mac)s IP: %(v4ip)s, netmask: %(v4mask)s, IPV6: %(v6ip)s, prefix: %(v6prefix)s status: %(status)s)" % self.__dict__

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__repr__() == other.__repr__()
        return False

# -- Read funcs


def readAllPortForwards(ignoreUUIDs=[]):
    return machine.readMachines() \
        .mapM(lambda m: readPortForwards(m) if m not in ignoreUUIDs else OK([])) \
        .map(flatten)


def readPortForwards(uuid):
    return vboxManager("showvminfo", "--machinereadable \"%s\"" % uuid) >> parsePortForwards


def readPortForward(uuid, name=None):
    return vboxManager("showvminfo", "--machinereadable \"%s\"" % uuid) \
        >> parsePortForwards \
        >> defer(filterPortForwards, name=name)


def readDHCPs():
    return vboxManager("list", "dhcpservers") \
        .bind(defer(_mapAsBlocks, func=parseDHCPBlock))


def readDHCP(interface):
    return readDHCPs() >> defer(filterDHCPs, interface=interface)


def readHostOnlyInterfaces():
    return vboxManager("list", "hostonlyifs") \
        .bind(defer(_mapAsBlocks, func=parseHostOnlyInterfaceBlock))


def readHostOnlyInterface(name):
    return readHostOnlyInterfaces() >> defer(filterHostOnlyInterfaces, name=name)


def filterPortForwards(ports, name):
    item = next((item for item in ports if item.name == name), None)
    return OK(item) if item else Fail(VirtualBoxObjectNotFound("Port Forward named \"%s\" was not found." % name))


def filterHostOnlyInterfaces(hoifs, name):
    item = next((item for item in hoifs if item.name == name), None)
    return OK(item) if item else Fail(VirtualBoxObjectNotFound("Host Only Interface \"%s\" was not found." % name))


def filterDHCPs(dhcps, interface):
    dhcp = next((dhcp for dhcp in dhcps if dhcp.interface == interface), None)
    return OK(dhcp) if dhcp else Fail(VirtualBoxObjectNotFound("DHCP for interface \"%s\" was not found." % interface))

# -- Parse funcs


def parseDHCPBlock(block):
    actions = (
        (r'^NetworkName:\s+HostInterfaceNetworking-(.+?)$', 'interface'),
        (r'^NetworkName:\s+(HostInterfaceNetworking-(.+?))$', 'networkName'),
        (r'^IP:\s+(.+?)$', 'gateway'),
        (r'^NetworkMask:\s+(.+?)$', 'netmask'),
        (r'^lowerIPAddress:\s+(.+?)$', 'lowerIP'),
        (r'^upperIPAddress:\s+(.+?)$', 'upperIP'),
        (r'Enabled:\s+(.+?)$', 'enabled')
    )
    return _extractClassFromBlock(block, actions, DHCP)


def parseHostOnlyInterfaceBlock(block):
    actions = (
        (r'^Name:\s+(.+?)$', 'name'),
        (r'^IPAddress:\s+(.+?)$', 'v4ip'),
        (r'^NetworkMask:\s+(.+?)$', 'v4mask'),
        (r'^IPV6Address:\s+(.+?)$', 'v6ip'),
        (r'^IPV6NetworkMaskPrefixLength:\s+(.+?)$', 'v6prefix'),
        (r'^Status:\s+(.+?)$', 'status'),
        (r'^VBoxNetworkName:\s+(.+?)$', 'netName'),
        (r'^HardwareAddress:\s+(.+?)$', 'mac'),
    )
    return _extractClassFromBlock(block, actions, HostOnlyInterface)


def parsePortForwards(vminfo):
    '''
    Parse Virtual Box machine info for forwarded ports.
    '''
    lines = vminfo.split(_vboxLineEnding())
    ports = []
    nic = None
    for line in lines:
        line = line.strip()

        # First match a nic from the info values
        nicmatch = re.match(r'^nic(\d+)=".+?"$', line)
        if nicmatch:
            nic = nicmatch.group(1)

        portmatch = re.match(
            r'^Forwarding.+?="(.+?),(.+?),(.*?),(.+?),(.*?),(.+?)"$', line)
        if portmatch:
            ports.append(PortForward(
                nic=nic,
                name=portmatch.group(1),
                proto=portmatch.group(2),
                hostIP=portmatch.group(3),
                hostPort=portmatch.group(4),
                guestIP=portmatch.group(5),
                guestPort=portmatch.group(6)
            ))

    return OK(ports)

# -- Actions


def clearPortForwards(uuid):
    return readPortForwards(uuid).bind(defer(removePortForwards, uuid=uuid))


def removePortForwards(ports, uuid):
    args = []
    for port in ports:
        args.append(port.getDeleteArg())

    if len(args) > 0:
        return vboxManager("modifyvm", "\"%s\" %s" % (uuid, " ".join(args)))
    else:
        return OK(None)


def addPortForwards(ports, uuid):
    args = []
    for port in ports:
        args.append(port.getCreateArg())

    if len(args) > 0:
        return vboxManager("modifyvm", "\"%s\" %s" % (uuid, " ".join(args)))
    else:
        return OK(None)


def addDHCP(hoif_name, gateway, netmask, lowerIP, upperIP):
    arg = "--ifname \"%s\" --ip \"%s\" --netmask \"%s\" --lowerip \"%s\" --upperip \"%s\" --enable" % (
        hoif_name, gateway, netmask, lowerIP, upperIP)
    return vboxManager("dhcpserver", "add %s" % arg) \
        .then(defer(enableDHCP, hoif_name=hoif_name))


def enableDHCP(hoif_name):
    arg = "--ifname \"%s\" --enable" % hoif_name
    return vboxManager("dhcpserver", "modify %s" % arg)


def disableDHCP(hoif_name):
    arg = "--ifname \"%s\" --disable" % hoif_name
    return vboxManager("dhcpserver", "modify %s" % arg)


def removeDHCP(hoif_name):
    return vboxManager("dhcpserver", "remove --ifname \"%s\"" % hoif_name)


def addHostOnlyInterface():
    def parseResult(r):
        match = re.match("^Interface '(.+?)' was successfully created", r)
        if match:
            return OK(match.group(1))
        else:
            return Fail(VirtualBoxError("Failed to create a new Host Only Interface"))
    return vboxManager("hostonlyif", "create").bind(parseResult)


def configureHostOnlyInterface(name, ip, netmask):
    net = _netFrom(ip + "/" + netmask)
    if net.isFail():
        return net

    net = net.getOK()
    # if net.version == 6:
    #  return vboxManager("hostonlyif", "ipconfig \"%s\" --ipv6 \"%s\" --netmasklengthv6 \"%s\"" % (name, net.ip, net.prefixlen))

    return vboxManager("hostonlyif", "ipconfig \"%s\" --ip \"%s\" --netmask \"%s\"" % (name, net.ip, net.netmask))


def removeHostOnlyInterface(hoif):
    return removeDHCP(hoif.name) \
        .catch(lambda x: OK(hoif)) \
        .then(defer(vboxManager, "hostonlyif", "remove \"%s\"" % hoif.name))

# -- Private helpers


def _ipFrom(ip):
    try:
        return OK(IPAddress(ip))
    except Exception as err:
        return Fail(err)


def _netFrom(net):
    try:
        return OK(IPNetwork(net))
    except Exception as err:
        return Fail(err)


def _mapAsBlocks(data, func):
    sep = _vboxLineEnding()
    sep = "%s%s" % (sep, sep)
    blocks = [x for x in data.strip().split(sep) if x]
    return OK(blocks).mapM(func)


def _extractClassFromBlock(block, actions, cls):
    lines = block.split(_vboxLineEnding())

    info = OrderedDict()
    for expr, field in actions:
        info[field] = None

    for line in lines:
        line = line.strip()
        for expr, field in actions:
            match = re.match(expr, line)
            if match:
                info[field] = match.group(1)
                break

    if info:
        return OK(cls(**info))
    else:
        return OK(None)



import sys
import os
import logging
import paramiko
from paramiko.py3compat import u
import socket
from time import time, sleep
from collections import namedtuple
from substance.monads import *
from substance.exceptions import *
from subprocess import check_output
from substance.termutils import (
    hasTermios,
    saveTerminalAttrs,
    restoreTerminalAttrs,
    setTerminalInteractive,
    getTerminalSize,
    decodeTerminalBytes
)

logger = logging.getLogger(__name__)

LinkResponse = namedtuple(
    'LinkResponse', ['link', 'stdin', 'stdout', 'stderr', 'code', 'cmd'])


class Link(object):

    def __init__(self, keyFile, keyFormat="RSA", useAgent=False):
        self.keyFile = keyFile
        self.keyFormat = keyFormat
        self.useAgent = useAgent
        self.key = None
        self.username = "substance"
        self.port = 22
        self.connected = False
        self.client = None
        self.sftp = None

    def waitForConnect(self, maxtries=200, timeout=60):

        start = time()
        tries = 0

        while not self.connected:
            logger.debug("Connect attempt %s..." % tries)

            def clearKeys(err):
                self.client.get_host_keys().clear()
                logger.debug("Clearing host keys...")
                return OK(None)

            state = self.connect()  \
                .catchError(paramiko.ssh_exception.BadHostKeyException, clearKeys) \
                .catchError(paramiko.ssh_exception.SSHException, lambda err: OK(None)) \
                .catchError(socket.error, lambda err: OK(None))
            if state.isFail():
                return state
            if time() - start > timeout:
                return Fail(LinkConnectTimeoutExceeded("Timeout exceeded"))
            if tries >= maxtries:
                return Fail(LinkRetriesExceeded("Max retries exceeded"))
            tries += 1
            sleep(0.8)

        end = time()
        logger.debug("Connect time: %ss" % (end - start))
        self.connected = True
        return OK(self)

    def connectEngine(self, engine):
        connectDetails = engine.getConnectInfo()
        self.hostname = connectDetails.get('hostname')
        self.port = connectDetails.get('port')
        logger.debug("Connecting to engine %s via %s:%s" %
                     (engine.name, self.hostname, self.port))
        return self.loadKey().then(self.waitForConnect)

    def connectHost(self, host, port):
        self.hostname = host
        self.port = port
        return self.loadKey().then(self.waitForConnect)

    def loadKey(self):
        try:
            if not self.key:
                logger.debug("Connecting using %s key: %s" %
                             (self.keyFormat, self.keyFile))
                if self.keyFormat == 'DSA':
                    self.key = paramiko.dsskey.DSSKey.from_private_key_file(
                        self.keyFile, password=None)
                elif self.keyFormat == 'ECSDA':
                    self.key = paramiko.ecdsakey.ECDSAKey.from_private_key_file(
                        self.keyFile, password=None)
                else:
                    self.key = paramiko.rsakey.RSAKey.from_private_key_file(
                        self.keyFile, password=None)

        except Exception as err:
            return Fail(err)

        return OK(self.key)

    def connect(self):
        try:
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            options = self.getConnectOptions()
            self.client.connect(**options)
            self.connected = True
            return OK(self)
        except Exception as err:
            return Fail(err)

    def getConnectOptions(self):
        return {
            'hostname': self.hostname,
            'port': self.port,
            'username': self.username,
            'pkey': self.key,
            'allow_agent': True,
            'look_for_keys': True,
            'timeout': 1,
            'banner_timeout': 1
        }

    def command(self, cmd, *args, **kwargs):
        return self.runCommand(cmd, *args, **kwargs)

    def streamCommand(self, cmd, *args, **kwargs):
        self.runCommand(cmd, stream=True)

    def runCommand(self, cmd, sudo=False, stream=False, capture=False, interactive=False, shell=False, *args, **kwargs):

        import select

        cmd = "sudo %s" % cmd if sudo is True else "%s" % cmd

        logger.debug("LINKCOMMAND: %s" % cmd)

        try:
            saveTerminalAttrs(sys.stdin)

            # Setup our connection channel with forwarding
            channel = None
            if shell:
                # Set terminal options
                channel = self.client.invoke_shell()
            else:
                channel = self.client.get_transport().open_session()
                channel.get_pty()

            if interactive:
                channel.settimeout(0.0)
                setTerminalInteractive(sys.stdin)

            forward = paramiko.agent.AgentRequestHandler(channel)

            # Setup our buffers
            bufsize = 1024
            stdin =  b''
            stdout = b''
            stderr = b''

            # Star the command stream
            if shell:
                channel.send(cmd + "\n", *args, **kwargs)
            else:
                channel.exec_command(cmd, *args, **kwargs)

            isAlive = True
            watchHandles = [channel]
            if interactive:
                watchHandles.append(sys.stdin)

            while isAlive:
                self.resizePTY(channel)
                r, w, e = select.select(watchHandles, [], [])

                if channel.exit_status_ready():
                    isAlive = False
                    break

                if channel in r and channel.recv_ready():
                    d = channel.recv(bufsize)
                    if len(d) == 0:
                        isAlive = False
                    else:
                        if stream:
                            sys.stdout.write(decodeTerminalBytes(d))
                            sys.stdout.flush()
                        if capture:
                            stdout += d

                if channel in r and channel.recv_stderr_ready():
                    d = channel.recv_stderr(bufsize)
                    if stream:
                        sys.stderr.write(decodeTerminalBytes(d))
                        sys.stderr.flush()
                    if capture:
                        stderr += d

                if sys.stdin in r and isAlive and interactive:
                    x = os.read(sys.stdin.fileno(), bufsize)
                    if len(x) == 0:
                        isAlive = False
                    else:
                        channel.send(decodeTerminalBytes(x))

            code = channel.recv_exit_status()
            stdout = decodeTerminalBytes(stdout)
            stdin = decodeTerminalBytes(stdin)
            stderr = decodeTerminalBytes(stderr)

            if code != 0:
                logger.debug("code %s, stdout: %s, stderr: %s" % (code, stdout, stderr))
                return Fail(LinkCommandError(cmd=cmd, message="An error occured when running command.", stdout=stdout, stderr=stderr, code=code, link=self))

            return OK(LinkResponse(link=self, cmd=cmd, stdin=stdin, stdout=stdout, stderr=stderr, code=code))
        # except Exception as err:
        #     logger.debug("%s" % err.stdout)
        #     return Fail(err)
        except KeyboardInterrupt as err:
            return Fail(LinkCommandError(cmd=cmd, message="User Interrupted.", stdout=stdout, stderr=stderr, code=1, link=self))
        finally:
            restoreTerminalAttrs(sys.stdin)

    def resizePTY(self, channel):
        tty_width, tty_height = getTerminalSize()
        try:
            channel.resize_pty(width=int(tty_width), height=int(tty_height))
        except paramiko.ssh_exception.SSHException:
            pass

    def getClient(self):
        return self.client

    def getSFTP(self):
        if not self.sftp:
            self.sftp = self.client.open_sftp()
        return self.sftp

    def closeSFTP(self):
        return self.sftp.close()

    def upload(self, localPath, remotePath):
        return Try.attempt(self._put, localPath, remotePath)

    def download(self, remotePath, localPath):
        return Try.attempt(self._get, remotePath, localPath)

    def runScript(self, scriptPath, sudo=False):
        scriptName = "/tmp/%s.sh" % time()

        return self.upload(scriptPath, scriptName) \
            .then(defer(self.runCommand, "chmod +x %s" % scriptName, sudo=True)) \
            .then(defer(self.runCommand, cmd=scriptName, sudo=sudo, stream=True))

    def interactive(self, cmd=None):
        if hasTermios:
            return Try.attempt(self._posixShell, cmd)
        else:
            return Try.attempt(self._windowsShell, cmd)

    def _posixShell(self, cmd=None):

        import select

        logger.debug("Opening interactive POSIX shell")
        sys.stdout.write('\r\n*** Begin interactive session.\r\n')
        saveTerminalAttrs(sys.stdin)

        channel = self.client.get_transport().open_session()
        forward = paramiko.agent.AgentRequestHandler(channel)

        channel.get_pty()
        channel.invoke_shell()

        if cmd:
            channel.send(cmd)
            channel.send("\n")

        try:
            setTerminalInteractive(sys.stdin)
            channel.settimeout(0.0)

            isAlive = True

            while isAlive:
                self.resizePTY(channel)
                r, w, e = select.select([channel, sys.stdin], [], [])
                if channel in r:
                    try:
                        x = channel.recv(1024)
                        if len(x) == 0:
                            isAlive = False
                        else:
                            sys.stdout.write(decodeTerminalBytes(x))
                            sys.stdout.flush()
                    except socket.timeout:
                        pass
                if sys.stdin in r and isAlive:
                    x = os.read(sys.stdin.fileno(), 1)
                    if len(x) == 0:
                        isAlive = False
                    else:
                        channel.send(x)
            channel.shutdown(2)

        finally:
            restoreTerminalAttrs(sys.stdin)
            sys.stdout.write('\r\n*** End of interactive session.\r\n')

    def _windowsShell(self, cmd=None):
        import threading

        sys.stdout.write(
            "Line-buffered terminal emulation. Press F6 or ^Z to send EOF.\r\n\r\n")

        def writeall(sock):
            while True:
                data = sock.recv(256)
                if not data:
                    sys.stdout.write('\r\n*** EOF ***\r\n\r\n')
                    sys.stdout.flush()
                    break
                sys.stdout.write(decodeTerminalBytes(data))
                sys.stdout.flush()

        writer = threading.Thread(target=writeall, args=(chan,))
        writer.start()

        try:
            while True:
                d = sys.stdin.read(1)
                if not d:
                    break
                chan.send(d)
        except EOFError:
            # user hit ^Z or F6
            pass

    def _put(self, localPath, remotePath):
        client = self.getSFTP()
        logger.debug("PUT %s %s" % (localPath, remotePath))
        return client.put(localPath, remotePath)

    def _get(self, remotePath, localPath):
        client = self.getSFTP()
        logger.debug("GET %s %s" % (remotePath, localPath))
        return client.get(remotePath, localPath)



from substance.monads import *
from substance.logs import *
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)


class Sshinfo(Command):

    def getUsage(self):
        return "substance sshinfo [ENGINE NAME]"

    def getHelpTitle(self):
        return "Obtain the ssh info configuration for connecting to an engine"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):

        name = self.getInputName()

        self.core.loadEngine(name) \
            .bind(Engine.loadConfigFile) \
            .bind(self.outputSSHInfo) \
            .catch(self.exitError)

    def outputSSHInfo(self, engine):
        host = engine.getSSHIP()
        port = engine.getSSHPort()

        sshconfig = """Host %(name)s
  HostName %(host)s
  User substance
  Port %(port)s
  UserKnownHostsFile /dev/null
  IdentityFile "%(keyfile)s"
  StrictHostKeyChecking no
  PasswordAuthentication no
  LogLevel FATAL
  ForwardAgent yes"""

        sshconfig = sshconfig % ({
            'name': engine.name,
            'host': host,
            'port': port,
            'keyfile': self.core.getInsecureKeyFile()
        })

        print(sshconfig)
        return OK(sshconfig)



import os
import logging
import shlex
from tabulate import tabulate

from substance.logs import *
from substance.exceptions import (InvalidOptionError)
from substance import Command
from substance.subenv import (SubenvAPI)

logger = logging.getLogger(__name__)


class Ls(Command):

    def getUsage(self):
        return "subenv ls [options]"

    def getHelpTitle(self):
        return "Print a list of substance project environments."

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        return self.api.ls() \
            .mapM(self.tabulateEnv) \
            .bind(self.tabulateEnvs) \
            .catch(self.exitError) \
            .bind(logger.info)

    def tabulateEnv(self, env):
        return OK([env.name, "X" if env.current else "", env.basePath, env.getLastAppliedDateTime('%B %d %Y, %H:%M:%S')])

    def tabulateEnvs(self, envs):
        headers = ["NAME", "CURRENT", "BASEPATH", "MODIFIED"]
        table = tabulate(envs, headers=headers, tablefmt="plain")
        return OK(table)



from substance import (Command, Engine)
from substance.constants import Syncher


class Sync(Command):
    def getShellOptions(self, optparser):
        optparser.add_option(
            "-u", "--up", dest="up", help="Synchronize UP to the engine only.", default=False, action="store_true")
        optparser.add_option("-d", "--down", dest="down",
                             help="Synchronize DOWN from the engine only.", default=False, action="store_true")
        optparser.add_option(
            "-i",
            "--ignore-archives",
            dest="ignorearchives",
            help="Ignore previous Unison archives and rebuild them (unison sync mode only)",
            default=False,
            action="store_true")
        return optparser

    def getUsage(self):
        return "substance sync [options] [SUBPATH]"

    def getHelpTitle(self):
        return "Synchronize & watch the devroot of the current engine"

    def getHelpDetails(self):
        return """
This process will run in the foreground and monitor local and remote file
system events to perform bidirectional file synchronization. When working
locally ; activate the sync command to keep your code up to date on the engine
or to receive artifact stored in the devroot in your local folders.

By default, this command syncs the entire engine home directory. You can
optionally pass a sub-path to the sync command. The sub-path is relative from
the current engine's home directory. If a valid sub-path is specified, then
syncing will only be active from within that sub-path, but the sub-path MUST
be valid on both sides of the sync (it will never be automatically created for
you).
"""

    def main(self):

        if self.getOption('up'):
            direction = Syncher.UP
        elif self.getOption('down'):
            direction = Syncher.DOWN
        else:
            direction = Syncher.BOTH

        path = ""
        if self.args:
            path = self.args[0]

        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(self.syncFolders, direction=direction, path=path) \
            .catch(self.exitError)

    def syncFolders(self, engine, direction=Syncher.BOTH, path=''):
        syncher = engine.getSyncher()
        if self.getOption('ignorearchives'):
            syncher.ignoreArchives = True
        return syncher.start(direction, path)



import os
import logging
import shlex

from substance.logs import *
from substance.exceptions import (InvalidOptionError)
from substance import Command
from substance.subenv import (SubenvAPI)


class Vars(Command):

    def getUsage(self):
        return "subenv vars [options] [VAR1 VAR2...?]"

    def getHelpTitle(self):
        return "Output the vars for an env on stdout, optionally filtered"

    def getShellOptions(self, optparser):
        optparser.add_option("-e", "--env", dest="env",
                             help="Specify an environment name")
        return optparser

    def main(self):
        name = self.readInputName().getOK()
        return self.api.vars(name, vars=self.getInputVars()) \
            .bind(self.printEnvVars) \
            .catch(self.exitError)

    def printEnvVars(self, vars=None):
        vars = {} if not vars else vars
        for k, v in vars.items():
            print(("%s=\"%s\"" % (k, v)))
        return OK(None)

    def getInputVars(self):
        if len(self.args) > 0:
            return self.args
        return None

    def readInputName(self):
        if self.getOption('env'):
            name = os.path.normpath(self.getOption('env'))
            return OK(name)
        else:
            return OK(None)



import re
import os
import platform
import logging

from substance.shell import Shell
from substance.monads import *
from substance.logs import *
from .exceptions import *
from substance.exceptions import *
from substance.path import (inner, outer)
from substance.platform import (isWSL, isWithinWindowsSubsystem)

VBOX_VERSION_MIN = (6, 0, 0)
VBOX_VERSION_CHECKED = None

logger = logging.getLogger(__name__)

def vboxManager(cmd, params=""):
    '''
    Invoke the VirtualBoxManager
    '''
    return assertVirtualBoxVersion().then(defer(vboxExec, cmd, params))

def getVBoxManageBin():
    return "VBoxManage.exe" if isWSL() else "VBoxManage"

def vboxExec(cmd, params=""):
    return Shell.procCommand("%s %s %s" % (getVBoxManageBin(), cmd, params)) \
        .bind(onVboxCommand) \
        .catch(onVboxError)


def onVboxCommand(sh):
    return OK(sh.get('stdout', ''))


def onVboxError(err):
    if isinstance(err, ShellCommandError):
        codeMatch = re.search(
            r'error: Details: code (VBOX_[A-Z_0-9]*)', err.stderr, re.M)
        code = codeMatch.group(1) if codeMatch else None
        return Fail(VirtualBoxError(message=err.stderr, code=code))


def readVersion():
    return vboxExec("--version").bind(parseVersion)


def parseVersion(vstring):
    fail = Fail(VirtualBoxVersionError(
        "Invalid version of VirtualBox installed : %s" % vstring))

    vstring = vstring.strip()
    if vstring is None or vstring == "":
        return fail

    parts = vstring.split('_')

    if len(parts) == 0:
        return fail

    return OK(parts[0].split("r")[0])


def assertVirtualBoxVersion():
    global VBOX_VERSION_CHECKED
    if VBOX_VERSION_CHECKED:
        return OK(VBOX_VERSION_CHECKED)
    else:
        return readVersion().bind(checkVersion).bind(saveVersion)


def saveVersion(vstring):
    global VBOX_VERSION_CHECKED
    VBOX_VERSION_CHECKED = vstring
    return OK(vstring)


def checkVersion(vstring):
    semver = vstring.split('.')
    semvercheck = tuple(1 for i in range(0, 3) if int(
        semver[i]) >= VBOX_VERSION_MIN[i])
    if semvercheck == (1, 1, 1):
        return OK(vstring)
    else:
        msg = "VirtualBox version %s and up is required. %s currently installed." % (
            VBOX_VERSION_MIN, vstring)
        return Fail(VirtualBoxVersionError(msg))


def _vboxLineEnding():
    if isWithinWindowsSubsystem():
        # Special case: os.linesep reports LF because of Cygwin/WSL but VBoxManage.exe will output CRLF because of Windows
        return "\r\n"
    return os.linesep



import logging
import tempfile
import os
from python_hosts import Hosts, HostsEntry
from shutil import copyfile
from substance.monads import *
from substance.shell import Shell

logger = logging.getLogger(__name__)


class SubHosts(Hosts):

    def __init__(self, path=None, tempFile=None):
        super(SubHosts, self).__init__(path)
        self.tempFile = tempFile

    def addEntry(self, name, destAddress):
        self.removeEntry(name)
        entry = HostsEntry(entry_type='ipv4',
                           address=destAddress, names=[name])
        self.add([entry], True)
        return entry

    def removeEntry(self, name):
        self.remove_all_matching(None, name)

    @staticmethod
    def register(name, destAddress):
        try:
            logger.info("Hosts register %s: %s" % (name, destAddress))
            hosts = SubHosts.checkoutFromSystem()
            entry = hosts.findEntryByAddress(address=destAddress)
            if entry is None:
                logger.info(
                    "'%s' Not found in local hosts file. Adding." % destAddress)
                entry = hosts.addEntry(name, destAddress)
                hosts.write()
                hosts.commitToSystem()

            if name not in entry.names:
                logger.info(
                    "'%s' Not found in local hosts file. Adding." % name)
                existing = hosts.findEntryByName(name=name)
                if existing:
                    existing.names.remove(name)
                entry.names.append(name)
                hosts.write()
                hosts.commitToSystem()
            else:
                logger.info("'%s' found in local hosts file." % name)

            return OK(None)
        except Exception as err:
            return Fail(err)

    @staticmethod
    def unregister(name):
        try:
            logger.info("Hosts unregister %s" % (name))
            hosts = SubHosts.checkoutFromSystem()
            if hosts.exists(names=[name]):
                hosts.removeEntry(name)
                hosts.write()
                hosts.commitToSystem()
            return OK(None)
        except Exception as err:
            return Fail(err)

    @staticmethod
    def checkoutFromSystem():
        tempFile = tempfile.NamedTemporaryFile(delete=False)
        tempFile.close()
        hostsPath = Hosts.determine_hosts_path()
        logger.debug("Checking out '%s' from system to '%s'" %
                     (hostsPath, tempFile.name))
        copyfile(hostsPath, tempFile.name)
        return SubHosts(tempFile.name, tempFile)

    def commitToSystem(self):
        hostsPath = self.determine_hosts_path()
        logger.debug("Commiting '%s' to system '%s'" %
                     (self.hosts_path, hostsPath))
        return Shell.call(["cp", self.hosts_path, self.determine_hosts_path()], shell=False, sudo=True) \
            .then(lambda: os.remove(self.tempFile.name))

    def findEntryByAddress(self, address):
        for entry in self.entries:
            entry_names = entry.names if entry.names is not None else []
            if address and address == entry.address:
                return entry
        return None

    def findEntryByName(self, name):
        for entry in self.entries:
            entry_names = entry.names if entry.names is not None else []
            if name in entry_names:
                return entry
        return None

    def exists(self, address=None, names=[]):
        for entry in self.entries:
            entry_names = entry.names if entry.names is not None else []
            if address and address == entry.address:
                return True
            if len(names) > 0:
                for name in names:
                    if name in entry_names:
                        return True
        return False



import os
import time
from collections import OrderedDict
from substance.monads import *
from substance.constants import *
from substance.utils import readDotEnv, writeToFile, makeSymlink, DictCompat
from substance.exceptions import (InvalidEnvError, InvalidOptionError)
from substance.config import (Config)
from substance import Shell
from substance.subenv.constants import (SPECDIR, ENVFILE, CODELINK, CONFFILE)
import jinja2
from functools import reduce

logger = logging.getLogger(__name__)


class SubenvSpec(object):
    def __init__(self, specPath, basePath, name=None, vars={}, lastApplied=None):
        self.specPath = specPath
        self.basePath = basePath
        self.envPath = None
        self.name = name
        self.envFiles = []
        self.overrides = vars
        self.vars = OrderedDict()
        self.lastApplied = lastApplied
        self.current = False
        self.struct = {'files': [], 'dirs': []}

    def setEnvPath(self, path):
        self.envPath = path
        self.config = Config(os.path.join(self.envPath, CONFFILE))

    @staticmethod
    def fromEnvPath(path):
        if not os.path.isdir(path):
            return Fail(InvalidOptionError("Specified path '%s' does not exist." % path))

        envPath = path
        name = os.path.basename(envPath)

        envVars = Try.attempt(readDotEnv, os.path.join(envPath, ENVFILE))
        if envVars.isFail():
            return envVars
        envVars = envVars.getOK()

        reserved = [
            'SUBENV_NAME',
            'SUBENV_LASTAPPLIED',
            'SUBENV_ENVPATH',
            'SUBENV_SPECPATH',
            'SUBENV_BASEPATH'
        ]
        vars = envVars.copy()
        for k in list(vars.keys()):
            if k in reserved:
                del vars[k]

        lastApplied = None
        if 'SUBENV_LASTAPPLIED' in envVars:
            lastApplied = envVars['SUBENV_LASTAPPLIED']

        env = SubenvSpec(envVars['SUBENV_SPECPATH'], envVars['SUBENV_BASEPATH'],
                         envVars['SUBENV_NAME'], vars, lastApplied)
        env.setEnvPath(envPath)
        return env

    @staticmethod
    def fromSpecPath(path, vars={}):
        if not os.path.isdir(path):
            return Fail(InvalidEnvError("Specified path '%s' does not exist." % path))

        if os.path.basename(path) == SPECDIR or not os.path.isdir(os.path.join(path, SPECDIR)):
            return Fail(InvalidOptionError("Invalid path specified. Please pass a path to a folder with a %s directory." % SPECDIR))
        specPath = os.path.join(path, SPECDIR)

        name = os.path.basename(path)

        return SubenvSpec(specPath, path, name, vars).scan()

    def getLastAppliedDateTime(self, fmt='%Y-%m-%d %H:%M:%S'):
        if self.lastApplied:
            return time.strftime(fmt, time.localtime(float(self.lastApplied)))
        return None

    def scan(self):
        return self.loadEnvVars(self.overrides) \
            .then(self.loadEnvStruct)

    def applyTo(self, envPath):
        self.setEnvPath(envPath)
        return self.clearEnv() \
            .then(self.applyDirs) \
            .then(self.applyFiles)  \
            .then(self.writeEnv) \
            .then(self.linkCode) \
            .then(self.assertConfig) \
            .then(self.applyScript) \
            .then(lambda: OK(self))

    def linkCode(self):
        return Try.sequence([
            Try.attempt(makeSymlink, self.basePath,
                        os.path.join(self.envPath, CODELINK), True),
            Try.attempt(makeSymlink, self.specPath,
                        os.path.join(self.envPath, SPECDIR), True)
        ])

    def clearEnv(self):
        if os.path.isfile(self.config.configFile):
            return Shell.rmFile(self.config.configFile)
        return OK(None)

    def writeEnv(self):
        dotenv = os.path.join(self.envPath, ENVFILE)
        logger.debug("Writing environment to: %s" % dotenv)
        envVars = OrderedDict(self.vars, **self.overrides)
        envVars.update({
            'SUBENV_NAME': self.name,
            'SUBENV_LASTAPPLIED': time.time(),
            'SUBENV_ENVPATH': self.envPath,
            'SUBENV_SPECPATH': self.specPath,
            'SUBENV_BASEPATH': self.basePath
        })

        env = "\n".join(["%s=\"%s\"" % (k, v) for k, v in envVars.items()])
        return Try.attempt(writeToFile, dotenv, env.encode())

    def assertConfig(self):
        if not os.path.isfile(self.config.configFile):
            return OK({})
        return self.config.loadConfigFile()

    def applyScript(self):
        commands = self.config.get('script', [])
        return Try.sequence(list(map(self.applyCommand,  commands)))

    def applyCommand(self, cmd):
        logger.info("Running environment command: %s" % cmd)
        return Shell.call(cmd, cwd=self.envPath, shell=True)

    def applyDirs(self):
        ops = [Shell.makeDirectory(self.envPath, 0o750)]
        for dir in self.struct['dirs']:
            sourceDir = os.path.join(self.specPath, dir)
            destDir = os.path.join(self.envPath, dir)
            mode = os.stat(sourceDir).st_mode & 0o777
            logger.debug("Creating directory '%s' mode: %s" % (dir, mode))
            ops.append(Shell.makeDirectory(destDir).bind(
                defer(Shell.chmod, mode=mode)))

        return Try.sequence(ops)

    def applyFiles(self):
        ops = []
        for file in self.struct['files']:
            fname, ext = os.path.splitext(file)
            source = os.path.join(self.specPath, file)
            dest = os.path.join(self.envPath, file)

            if fname == ENVFILE:
                continue
            elif ext == '.jinja':
                logger.debug("Rendering '%s' to %s" % (file, dest))
                dest = os.path.splitext(dest)[0]
                ops.append(self.renderFile(file, dest, self.vars))
            else:
                logger.debug("Copying '%s' to %s" % (file, dest))
                ops.append(Shell.copyFile(os.path.join(
                    self.specPath, file), os.path.join(self.envPath, file)))

        return Try.sequence(ops)

    def getEnvVars(self):
        envVars = self.overrides.copy()
        envVars.update({
            'SUBENV_NAME': self.name,
            'SUBENV_LASTAPPLIED': self.lastApplied,
            'SUBENV_ENVPATH': self.envPath,
            'SUBENV_SPECPATH': self.specPath,
            'SUBENV_BASEPATH': self.basePath
        })
        return envVars

    def renderFile(self, source, dest, vars={}):
        try:
            tplEnv = jinja2.Environment(
                loader=jinja2.FileSystemLoader(self.specPath))
            tpl = tplEnv.get_template(source)

            tplVars = self.getEnvVars()
            tplVars.update({'subenv': self})
            tplVars.update({'SUBENV_VARS': DictCompat({**self.getEnvVars()})})

            with open(dest, 'wb') as fh:
                fh.write(tpl.render(**tplVars).encode())
            return OK(None)
        except Exception as err:
            return Fail(err)

    def loadEnvStruct(self):
        return self.scanEnvStruct().bind(self.setEnvStruct)

    def scanEnvStruct(self):
        struct = {'dirs': [], 'files': []}
        for root, dirs, files in os.walk(self.specPath):
            relPath = os.path.relpath(
                root, self.specPath).strip('./').strip('/')
            for dir in dirs:
                struct['dirs'].append(os.path.join(relPath, dir))
            for file in files:
                struct['files'].append(os.path.join(relPath, file))
        return OK(struct)

    def setEnvStruct(self, struct):
        self.struct = struct
        return OK(self)

    def setEnvVars(self, e={}):
        self.overrides = e
        return OK(self)

    def loadEnvVars(self, env={}):
        return self.readEnvFiles() \
            .map(lambda e: dict(e, **env)) \
            .bind(self.setEnvVars)

    def readEnvFiles(self):
        specEnvFile = os.path.join(self.specPath, ENVFILE)
        if os.path.isfile(specEnvFile):
            self.envFiles.append(specEnvFile)

        baseEnvFile = os.path.join(self.basePath, ENVFILE)
        if os.path.isfile(baseEnvFile):
            self.envFiles.append(baseEnvFile)

        list(map(lambda x: logger.info("Loading dotenv file: '%s'" % x), self.envFiles))
        return Try.sequence(list(map(Try.attemptDeferred(readDotEnv), self.envFiles)))  \
            .map(lambda envs: reduce(lambda acc, x: dict(acc, **x), envs, {}))

    def __repr__(self):
        return "SubEnvSpec(%(name)s) spec:%(specPath)s base:%(basePath)s envFile:%(envFiles)s vars:%(vars)s files:%(struct)s" % self.__dict__



# -*- coding: utf-8 -*-
# $Id$


class SubstanceError(Exception):
    '''
    Base class to all substance exceptions
    '''
    errorLabel = ""

    def __init__(self, message=''):
        super(SubstanceError, self).__init__(message)
        self.errorLabel = message


class InvalidCommandError(SubstanceError):
    '''
    Raised when an invalid command is passed to CLI
    '''


class InvalidOptionError(SubstanceError):
    '''
    Raised when an invalid option is passed to CLI
    '''


class ShellCommandError(SubstanceError):
    def __init__(self, code=None, message='', stdout='', stderr='', cmd=''):
        super(SubstanceError, self).__init__(message)
        self.code = code
        self.stdout = stdout
        self.stderr = stderr
        self.cmd = cmd


class UserInterruptError(SubstanceError):
    '''
    Raised when a shell command is interrupted by the user.
    '''


class FileSystemError(SubstanceError):
    '''
    Raised when substance fails to read or write to the filesystem
    '''


class FileDoesNotExist(SubstanceError):
    '''
    Raised when a file expected to be there does not exist.
    '''


class ConfigSyntaxError(SubstanceError):
    '''
    Raised when invalid syntax was found in a configuration file.
    '''


class ConfigValidationError(SubstanceError):
    '''
    Raised when configuration does not validate.
    '''


class EngineError(SubstanceError):
    '''
    Engine error class
    '''
    engine = None

    def __init__(self, message='', engine=None):
        super(SubstanceError, self).__init__(message)
        self.engine = engine


class EngineExistsError(EngineError):
    '''
    Raised when an engine with the same name exists.
    '''


class EngineNotFoundError(EngineError):
    '''
    Raised when a specified engine was not found
    '''


class EngineNotProvisioned(EngineError):
    '''
    Engine VM is not provided. (ex. does not exist)
    '''


class EngineProvisioned(EngineError):
    '''
    Raised when an engine is provisioned and remove call is initiated.
    '''


class EngineAlreadyProvisioned(EngineError):
    '''
    Engine VM is already provided.
    '''


class EngineAlreadyRunning(EngineError):
    '''
    Engine VM is already running.
    '''


class EngineNotRunning(EngineError):
    '''
    Engine VM is not running.
    '''


class EnvNotDefinedError(EngineError):
    '''
    Raised when an env is not defined on an engine
    '''


class InvalidOrchestrator(EngineError):
    '''
    Raised when a subenv configures an unknown orchestrator
    '''


class EnvNotFoundError(EngineError):
    '''
    Raised when an env is not found
    '''


class SubstanceDriverError(SubstanceError):
    '''
    Driver-specific raised error
    '''


class MachineDoesNotExist(SubstanceDriverError):
    '''
    Machine does not exist
    '''


class LinkConnectTimeoutExceeded(SubstanceError):
    '''
    Link connect timeout exceeded
    '''


class LinkTimeoutExceeded(SubstanceError):
    '''
    Link timeout exceeded
    '''


class LinkRetriesExceeded(SubstanceError):
    '''
    Link retries exceeded
    '''


class LinkCommandError(ShellCommandError):
    '''
    Link command error
    '''

    def __init__(self, code=None, message='', stdout='', stderr='', cmd='', link=None):
        super(SubstanceError, self).__init__(message)
        self.code = code
        self.stdout = stdout
        self.stderr = stderr
        self.cmd = cmd
        self.link = link


class BoxError(SubstanceError):
    '''
    Box error category
    '''


class InvalidBoxName(BoxError):
    '''
    Name of box was not formatted correctly
    '''


class BoxArchiveHashMismatch(BoxError):
    '''
    Raised when a box archive hash does not match the registry.
    '''


class SubenvError(SubstanceError):
    '''
    Subenv errors
    '''


class InvalidEnvError(SubenvError):
    '''
    Raised when an invalid env is specified
    '''



import logging
from substance.logs import *
from substance.monads import *
from substance import (Command, Box, Shell)
from substance.exceptions import (SubstanceError)


class Pull(Command):

    def getUsage(self):
        return "substance box pull [BOXSTRING]"

    def getHelpTitle(self):
        return "Download a box"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        name = self.getInputName()
        return self.core.readBox(name) \
            .bind(self.core.pullBox) \
            .then(dinfo("Box \"%s\" has been pulled.", name)) \
            .catch(self.exitError)

    def getInputName(self):
        if len(self.args) <= 0:
            return self.exitError("Please specify a box string.")
        name = self.args[0]
        return name



from substance.monads import *
from substance.logs import *
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)


class Ssh(Command):

    def getUsage(self):
        return "substance ssh [ENGINE NAME]"

    def getHelpTitle(self):
        return "Connect using SSH to the engine's virtual machine."

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        name = self.getInputName()

        self.core.loadEngine(name) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.shell) \
            .catch(self.exitError)



import unittest
from substance.monads import *
from collections import namedtuple


class TestMonads(unittest.TestCase):

    def testJust(self):
        self.assertTrue(Just(5) == Just(5))
        self.assertFalse(Just(5) == Just(3))
        self.assertEqual(Just(10), Just(5).map(lambda i: i * 2))
        self.assertEqual(Just(10), Just(5).bind(lambda x: Just(x * 2)))

    def testOperators(self):
        base = Just(10)

        testRshift = Just(5) >> (lambda x: Just(x * 2))
        self.assertEqual(base, testRshift)

        testLshift = Just(5)
        testLshift <<= (lambda x: Just(x * 2))
        self.assertEqual(base, testLshift)

        self.assertTrue(isinstance(Maybe.of(True), Just))
        self.assertTrue(isinstance(Maybe.of(None), Nothing))
        self.assertTrue(isinstance(Maybe.of(), Nothing))
        self.assertTrue(isinstance(Maybe.of(False), Just))
        self.assertTrue(isinstance(Maybe.of(""), Just))
        self.assertTrue(isinstance(Maybe.of(1), Just))

    def testNothing(self):
        self.assertTrue(Nothing() == Nothing())
        self.assertFalse(Nothing() == Just(5))
        self.assertEqual(Nothing(), Nothing().map(lambda x: x * 100))
        self.assertEqual(Nothing(), Nothing().bind(lambda x: Just(x * 2)))

    def testMaybe(self):
        maybe = Just(5).map(lambda x: x * 2).bind(lambda x: Nothing()).map(lambda x: x * 2)
        self.assertEqual(Nothing(), maybe)

        maybe2 = Just(5).map(lambda x: x * 2).bind(lambda x: Just(x + 10)).map(lambda x: x * 2)
        self.assertEqual(Just(40), maybe2)

    def testRight(self):
        self.assertTrue(Right(5) == Right(5))
        self.assertFalse(Right(5) == Right(10))
        self.assertEqual(Right(10), Right(5).map(lambda x: x * 2))
        self.assertEqual(Right(10), Right(5).bind(lambda x: Right(x * 2)))
        self.assertEqual("%s" % Right(10), "Right(10)")

    def testLeft(self):
        self.assertTrue(Left(5) == Left(5))
        self.assertFalse(Left(5) == Left(10))
        self.assertEqual(Left(5), Left(5).map(lambda x: x * 2))
        self.assertEqual(Left(5), Left(5).bind(lambda x: Left(x * 2)))
        self.assertEqual("%s" % Left(5), "Left(5)")

    def testEither(self):
        either = Right(5).map(lambda x: x * 2).bind(lambda x: Left("Did not work")).map(lambda x: x * 2)
        self.assertEqual(Left("Did not work"), either)
        either2 = Right(5).map(lambda x: x * 2).bind(lambda x: Right(x + 10)).map(lambda x: x * 2)
        self.assertEqual(Right(40), either2)

    def testOK(self):
        self.assertTrue(OK(5) == OK(5))
        self.assertFalse(OK(5) == OK(10))
        self.assertEqual(OK(10), OK(5).map(lambda x: x * 2))
        self.assertEqual(OK(10), OK(5).bind(lambda x: OK(x * 2)))
        self.assertEqual("%s" % OK(5), "OK(5)")

    def testFail(self):
        valError = ValueError()
        synError = SyntaxError()
        self.assertTrue(Fail(valError) == Fail(valError))
        self.assertFalse(Fail(valError) == Fail(synError))
        self.assertEqual(Fail(valError), Fail(valError).map(lambda x: x * 2))
        self.assertEqual(Fail(valError), Fail(valError).bind(lambda x: OK(x * 2)))
        self.assertEqual("%s" % Fail(valError), "Fail(ValueError())")
        self.assertEqual(Fail(valError), Fail(valError).catch(lambda x: 1337))

    def testTry(self):
        valError = ValueError()
        synError = SyntaxError()

        t3y = OK(10).map(lambda x: x * 2).bind(lambda x: Fail(valError)).map(lambda x: x * 2)
        self.assertEqual(Fail(valError), t3y)

        t3y2 = OK(10).map(lambda x: x * 2).bind(lambda x: OK(50)).map(lambda x: x * 2)
        self.assertEqual(OK(100), t3y2)

    def testTryAttempt(self):

        t3r3 = Try.attempt(lambda: "Foo")
        self.assertEqual(OK("Foo"), t3r3)

        t3r3 = Try.attempt(Try.raiseError)
        self.assertEqual(t3r3.__class__.__name__, "Fail")

    def testTryThen(self):
        valError = ValueError()
        synError = SyntaxError()
        attempt = (OK(10)
                   .then(lambda: OK(20))
                   .then(lambda: OK(30))
                   .then(lambda: OK(40)))

        self.assertEqual(OK(40), attempt)

        attempt2 = (OK(10)
                    .then(lambda: OK(20))
                    .then(lambda: Fail(synError))
                    .then(lambda: OK(40)))

        self.assertEqual(Fail(synError), attempt2)

        attemptRecover = (OK(10)
                          .then(lambda: OK(20))
                          .then(lambda: Fail(valError))
                          .then(lambda: OK(40))
                          .catch(lambda err: OK("Recovered")))
        self.assertEqual(OK("Recovered"), attemptRecover)

        attemptFail = (OK(10)
                       .then(lambda: OK(20))
                       .then(lambda: Fail(valError))
                       .then(lambda: OK(40))
                       .catch(lambda err: "foo"))

        self.assertEqual(Fail(valError), attemptFail)

        attemptReraise = (OK(10)
                          .then(lambda: OK(20))
                          .then(lambda: Fail(valError))
                          .then(lambda: OK(40))
                          .catch(lambda err: Fail(synError)))

        self.assertEqual(Fail(synError), attemptReraise)

        attemptCatchOK = (OK(10)
                          .then(lambda: OK(20))
                          .then(lambda: OK(40))
                          .catch(lambda err: OK("123")))

        self.assertEqual(OK(40), attemptCatchOK)

    def testTryThenIfBool(self):
        valError = ValueError()
        self.assertEqual(OK(10).thenIfBool(lambda *x: OK(20), lambda *x: Fail(valError)), Fail(valError))
        self.assertEqual(OK(True).thenIfBool(lambda *x: OK(20), lambda *x: Fail(valError)), OK(20))
        self.assertEqual(OK(False).thenIfBool(lambda *x: OK(20), lambda *x: Fail(valError)), Fail(valError))
        self.assertEqual(Fail(valError).thenIfBool(lambda *x: OK(20), lambda *x: OK(20)), Fail(valError))
        self.assertEqual(OK(True).thenIfTrue(lambda *x: OK(10)), OK(10))
        self.assertEqual(OK(True).thenIfFalse(lambda *x: OK(10)), OK(True))
        self.assertEqual(OK(False).thenIfTrue(lambda *x: OK(10)), OK(False))
        self.assertEqual(OK(False).thenIfFalse(lambda *x: OK(10)), OK(10))

    def testTryThenIfNone(self):
        self.assertEqual(OK(10).thenIfNone(lambda *x: OK(20)), OK(10))
        self.assertEqual(OK(None).thenIfNone(lambda *x: OK(20)), OK(20))

    def testTryBindIf(self):
        valError = ValueError()
        self.assertEqual(OK(True).bindIf(lambda *x: OK(10), lambda *x: Fail(valError)), OK(10))
        self.assertEqual(OK(False).bindIf(lambda *x: OK(10), lambda *x: Fail(valError)), Fail(valError))
        self.assertEqual(OK(True).bindIfTrue(lambda *x: OK(10)), OK(10))
        self.assertEqual(OK(True).bindIfFalse(lambda *x: OK(10)), OK(True))
        self.assertEqual(OK(False).bindIfTrue(lambda *x: OK(10)), OK(False))
        self.assertEqual(OK(False).bindIfFalse(lambda *x: OK(10)), OK(10))

    def testTryCatchError(self):
        valError = ValueError()
        synError = SyntaxError()
        self.assertEqual(OK(True).catchError(ValueError, lambda *x: OK(10)), OK(True))
        self.assertEqual(OK(False).catchError(ValueError, lambda *x: OK(10)), OK(False))
        self.assertEqual(Fail(synError).catchError(ValueError, lambda e: OK(10)), Fail(synError))
        self.assertEqual(Fail(synError).catchError(SyntaxError, lambda e: OK(10)), OK(10))

    def testFunctions(self):
        self.assertEqual(defer(lambda x, y: x + y, 1, 2)(), 3)
        self.assertEqual(defer(lambda x, y: x + y)(1, 2), 3)
        valError = ValueError()
        self.assertEqual(failWith(valError)(), Fail(valError))
        self.assertEqual(failWith(valError)(1, 2, 3, 4), Fail(valError))

        def fa(x):
            return x + 2

        def fb(x):
            return x + 1

        def fc(acc, x):
            return acc + x

        self.assertEqual(compose(fa, fb)(1), fa(fb(1)))
        self.assertEqual(fold(fc, [1, 2, 3, 4], 0), 10)

    def testMapM(self):
        self.assertEqual(mapM(Try, lambda x: OK(x + x), [1, 2, 3, 4]), OK([2, 4, 6, 8]))

        testList = []
        self.assertEqual(mapM_(Try, lambda x: OK(testList.append(x + x)), [1, 2, 3, 4]), OK([1, 2, 3, 4]))
        self.assertEqual([2, 4, 6, 8], testList)

        self.assertEqual(OK([1, 2, 3, 4]).mapM(lambda x: OK(x + x)), OK([2, 4, 6, 8]))

        testList = []
        self.assertEqual(OK([1, 2, 3, 4]).mapM_(lambda x: OK(testList.append(x + x))), OK([1, 2, 3, 4]))
        self.assertEqual([2, 4, 6, 8], testList)

    def testSequence(self):
        self.assertEqual(sequence(Try, [OK(1), OK(2), OK(3), OK(4)]), OK([1, 2, 3, 4]))
        self.assertIsInstance(sequence(Try, [OK(1), OK(2), OK(3), Fail(ValueError())]), Fail)

        self.assertEqual(Try.sequence([OK(1), OK(2), OK(3)]), OK([1, 2, 3]))
        self.assertIsInstance(Try.sequence([OK(1), Fail(ValueError()), OK(3)]), Fail)

    def testUnshiftM(self):
        self.assertEqual(unshiftM(Try, OK([2, 3]), OK(1)), OK([1, 2, 3]))

    def testFlatten(self):
        self.assertEqual(['a', 'b', 'c', 'd', 'e', 'f'], flatten([['a', 'b'], ['c', 'd'], ['e', 'f']]))



import logging
from substance.logs import *
from substance.monads import *
from substance import (Command, Box, Shell)
from substance.exceptions import (SubstanceError)


class Delete(Command):

    def getUsage(self):
        return "substance box delete [BOXSTRING]"

    def getHelpTitle(self):
        return "Permanently delete a box."

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        name = self.getInputName()
        self.ask("You are about to delete box \"%s\"" % name) \
            .then(defer(self.core.readBox, name)) \
            .bind(self.core.removeBox) \
            .then(dinfo("Box \"%s\" has been deleted.", name)) \
            .catch(self.exitError)

    def getInputName(self):
        if len(self.args) <= 0:
            return self.exitError("Please specify a box string.")

        name = self.args[0]
        return name
